import numpy as np
import scipy
import scipy.sparse as sp

import openmdao.api as om

from dymos.transcriptions.grid_data import GridData
from dymos.utils.misc import get_rate_units
from dymos._options import options as dymos_options
from dymos.utils.birkhoff import birkhoff_matrix


class PicardUpdateComp(om.ExplicitComponent):
    """
    Class definition for the PicardUpdateComp.

    Given the initial state values (for forward propagation) or the final state values
    (for backward propagation), compute the next state value for picard iteration
    using a NonlinearBlockGS solver.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so we can determine shape and units.

        Parameters
        ----------
        phase : Phase
            The phase object that contains this collocation comp.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']
        num_segs = gd.num_segments
        time_units = self.options['time_units']
        state_options = self.options['state_options']

        B_blocks = []

        self._seg_repeats = gd.subset_num_nodes_per_segment['all']

        start_idx = 0
        for i in range(num_segs):
            nnps_i = gd.subset_num_nodes_per_segment['all'][i]
            tau_i = gd.node_stau[start_idx: start_idx + nnps_i]
            w_i = gd.node_weight[start_idx: start_idx + nnps_i]

            B_i = birkhoff_matrix(tau_i, w_i, grid_type=gd.grid_type)
            B_blocks.append(B_i)
            start_idx += nnps_i

        self._B = scipy.sparse.block_diag(B_blocks, format='csr')

        self.add_input('dt_dstau', units=self.options['time_units'], shape=(num_nodes,))

        self.var_names = var_names = {}
        for state_name, options in state_options.items():
            var_names[state_name] = {
                'f_computed': f'f_computed:{state_name}',
                'x_0': f'seg_initial_states:{state_name}',
                'x_f': f'seg_final_states:{state_name}',
                'x_a': f'initial_states:{state_name}',
                'x_b': f'final_states:{state_name}',
                'x_hat': f'states:{state_name}'
            }

        for state_name, options in state_options.items():
            shape = options['shape']
            size = np.prod(shape)
            units = options['units']

            rate_units = get_rate_units(units, time_units)
            var_names = self.var_names[state_name]

            rate_source = options['rate_source']
            rate_source_type = phase.classify_var(rate_source)

            if rate_source_type == 'state':
                var_names['f_computed'] = f'state_val:{rate_source}'
            else:
                self.add_input(
                    name=var_names['f_computed'],
                    shape=(num_nodes,) + shape,
                    desc=f'Computed derivative of state {state_name} at the polynomial nodes',
                    units=rate_units)

            self.add_output(
                name=var_names['x_hat'],
                shape=(num_nodes,) + shape,
                units=units
            )

            if options['solve_segments'] == 'forward':
                self.add_input(
                    name=var_names['x_0'],
                    shape=(num_segs,) + shape,
                    desc=f'Initial value of state {state_name} in each segment',
                    units=units
                )
                self.add_output(
                    name=var_names['x_b'],
                    shape=(1,) + shape,
                    desc=f'Final value of state {state_name} in the phase',
                    units=units
                )
                # The first row of these matrices are zero.
                rs = np.repeat(np.arange(1, num_nodes, dtype=int), num_nodes)
                cs = np.tile(np.arange(num_nodes, dtype=int), num_nodes - 1)
                self.declare_partials(of=var_names['x_hat'],
                                      wrt='dt_dstau',
                                      rows=rs, cols=cs)
                self.declare_partials(of=var_names['x_hat'],
                                      wrt=var_names['f_computed'],
                                      rows=rs, cols=cs)
                self.declare_partials(of=var_names['x_b'],
                                      wrt='dt_dstau')
                self.declare_partials(of=var_names['x_b'],
                                      wrt=var_names['f_computed'])

                rs = np.arange(num_nodes, dtype=int)
                cs = np.repeat(np.arange(num_segs, dtype=int), self._seg_repeats)
                self.declare_partials(of=var_names['x_hat'],
                                      wrt=var_names['x_0'],
                                      rows=rs, cols=cs, val=1.0)
                template = sp.lil_array((1, num_segs), dtype=int)
                template[0, -1] = 1
                template = sp.kron(template.tocsr(), sp.eye(size, dtype=int))
                rs, cs = template.nonzero()
                self.declare_partials(of=var_names['x_b'],
                                      wrt=var_names['x_0'],
                                      rows=rs, cols=cs, val=1.0)

            elif options['solve_segments'] == 'backward':
                self.add_input(name=var_names['x_f'],
                               shape=(num_segs,) + shape,
                               desc=f'Final value of state {state_name} in each segment',
                               units=units)
                self.add_output(name=var_names['x_a'],
                                shape=(1,) + shape,
                                desc=f'Initial value of state {state_name} in the phase',
                                units=units)
                rs = np.repeat(np.arange(num_nodes - 1, dtype=int), num_nodes)
                cs = np.tile(np.arange(num_nodes, dtype=int), num_nodes - 1)
                self.declare_partials(of=var_names['x_hat'], wrt='dt_dstau',
                                      rows=rs, cols=cs)
                self.declare_partials(of=var_names['x_hat'], wrt=var_names['f_computed'],
                                      rows=rs, cols=cs)
                self.declare_partials(of=var_names['x_a'],
                                      wrt='dt_dstau')
                self.declare_partials(of=var_names['x_a'],
                                      wrt=var_names['f_computed'])
                rs = np.arange(num_nodes, dtype=int)
                cs = np.repeat(np.arange(num_segs, dtype=int), self._seg_repeats)
                self.declare_partials(of=var_names['x_hat'],
                                      wrt=var_names['x_f'],
                                      rows=rs, cols=cs, val=1.0)
                template = sp.lil_array((1, num_segs), dtype=int)
                template[0, 0] = 1
                template = sp.kron(template.tocsr(), sp.eye(size, dtype=int))
                rs, cs = template.nonzero()
                self.declare_partials(of=var_names['x_a'], wrt=var_names['x_f'],
                                      rows=rs, cols=cs, val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        discrete_outputs : dict or None
            If not None, dict containing discrete output values.
        """
        dt_dstau = np.atleast_2d(inputs['dt_dstau']).T
        nn = self.options['grid_data'].num_nodes

        for state_name, options in self.options['state_options'].items():
            var_names = self.var_names[state_name]

            # Multiplication by B results in the integral in segment tau space,
            # so we need to convert f from being dx/dt to dx/dstau.
            f = np.einsum('i...,i...->i...', inputs[var_names['f_computed']], dt_dstau)
            f_flat = f.reshape(nn, -1)

            if options['solve_segments'] == 'forward':
                x_0 = inputs[var_names['x_0']]
                x_0_repeated = np.repeat(x_0, self._seg_repeats, axis=0)
                outputs[var_names['x_hat']] = x_0_repeated + (self._B @ f_flat).reshape(f.shape)
                outputs[var_names['x_b']][...] = outputs[var_names['x_hat']][-1, ...]

            elif options['solve_segments'] == 'backward':
                x_f = inputs[var_names['x_f']]
                x_f_repeated = np.repeat(x_f, self._seg_repeats, axis=0)
                outputs[var_names['x_hat']] = x_f_repeated - (self._B[::-1, ...] @ f_flat[::-1, ...]).reshape(f.shape)
                outputs[var_names['x_a']][...] = outputs[var_names['x_hat']][0, ...]

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        """
        Compute component partials.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : dict
            Partial values to be returend in partials[of, wrt] format.
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']

        dt_dstau = np.atleast_2d(inputs['dt_dstau']).T

        for state_name, options in self.options['state_options'].items():
            var_names = self.var_names[state_name]

            x_a_name = var_names['x_a']
            x_b_name = var_names['x_b']
            x_name = var_names['x_hat']
            f_name = var_names['f_computed']

            # Multiplication by B results in the integral in tau space,
            # so we need to convert f from being dx/dt to dx/dtau.
            f_t = inputs[var_names['f_computed']]

            if options['solve_segments'] == 'forward':
                partials[x_name, f_name] = (self._B.multiply(dt_dstau.T)).todense()[1:, ...].ravel()
                partials[x_b_name, f_name] = partials[x_name, f_name][-num_nodes:]
                partials[x_name, 'dt_dstau'] = (self._B.multiply(f_t.T)).todense()[1:, ...].ravel()
                partials[x_b_name, 'dt_dstau'] = partials[x_name, 'dt_dstau'][-num_nodes:]

            elif options['solve_segments'] == 'backward':
                B_flip = self._B[::-1, ...]
                dt_dstau_flip = dt_dstau[::-1, ...]

                partials[x_name, f_name] = -(B_flip.multiply(dt_dstau_flip.T)).todense()[:-1, ::-1].ravel()
                partials[x_a_name, f_name] = partials[x_name, f_name][:num_nodes]
                partials[x_name, 'dt_dstau'] = -(B_flip.multiply(f_t[::-1, ...].T)).todense()[:-1, ::-1].ravel()
                partials[x_a_name, 'dt_dstau'] = partials[x_name, 'dt_dstau'][:num_nodes]



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from dymos.utils.hermite import hermite_matrices


class TestHermiteMatrices(unittest.TestCase):

    def test_quadratic(self):

        # Interpolate with values and rates provided at [-1, 1] in tau space
        tau_given = [-1.0, 1.0]
        tau_eval = np.linspace(-1, 1, 100)

        # In time space use the boundaries [-2, 2]
        dt_dtau = 4.0 / 2.0

        # Provide values for y = t**2 and its time-derivative
        y_given = [4.0, 4.0]
        ydot_given = [-4.0, 4.0]

        # Get the hermite matrices.
        Ai, Bi, Ad, Bd = hermite_matrices(tau_given, tau_eval)

        # Interpolate y and ydot at tau_eval points in tau space.
        y_i = np.dot(Ai, y_given) + dt_dtau * np.dot(Bi, ydot_given)
        ydot_i = (1.0 / dt_dtau) * np.dot(Ad, y_given) + np.dot(Bd, ydot_given)

        # Compute our function as a point of comparison.
        y_computed = (tau_eval * dt_dtau)**2
        ydot_computed = 2.0 * (tau_eval * dt_dtau)

        # Check results
        assert_almost_equal(y_i, y_computed)
        assert_almost_equal(ydot_i, ydot_computed)

    def test_cubic(self):

        # Interpolate with values and rates provided at [-1, 1] in tau space
        tau_given = [-1.0, 0.0, 1.0]
        tau_eval = np.linspace(-1, 1, 101)

        # In time space use the boundaries [-2, 2]
        dt_dtau = 4.0 / 2.0

        # Provide values for y = t**2 and its time-derivative
        y_given = [-8.0, 0.0, 8.0]
        ydot_given = [12.0, 0.0, 12.0]

        # Get the hermite matrices.
        Ai, Bi, Ad, Bd = hermite_matrices(tau_given, tau_eval)

        # Interpolate y and ydot at tau_eval points in tau space.
        y_i = np.dot(Ai, y_given) + dt_dtau * np.dot(Bi, ydot_given)
        ydot_i = (1.0 / dt_dtau) * np.dot(Ad, y_given) + np.dot(Bd, ydot_given)

        # Compute our function as a point of comparison.
        y_computed = (tau_eval * dt_dtau)**3
        ydot_computed = 3.0 * (tau_eval * dt_dtau)**2

        # Check results
        assert_almost_equal(y_i, y_computed)
        assert_almost_equal(ydot_i, ydot_computed)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import unittest

import openmdao.api as om

import dymos as dm


class MyODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('polycon', shape=(nn,))
        self.add_input('con', shape=(nn,))
        self.add_input('state', shape=(nn,))
        self.add_output('out', shape=(nn,))
        self.add_output('state_deriv', shape=(nn,))

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        pass


class TestControlInterpComp(unittest.TestCase):

    def test_poly_con_bug(self):
        prob = om.Problem()

        phase = dm.Phase(ode_class=MyODE,
                         transcription=dm.GaussLobatto(num_segments=3))

        phase.add_state('state', rate_source='state_deriv')

        phase.add_control('polycon', order=1, val=0.0, opt=True, control_type='polynomial')
        phase.add_control('con', val=0.0, opt=True)

        phase.add_objective('out', loc='final', ref=1000.0)

        traj = dm.Trajectory()
        prob.model.add_subsystem('traj', traj)
        traj.add_phase('phase', phase)

        # Make sure we raise no exceptions.
        prob.setup()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om


class RobotArmODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('arm_length', types=float, default=5.0)

    def setup(self):
        nn = self.options['num_nodes']
        L = self.options['arm_length']

        # inputs: 6 states and 3 controls
        self.add_input(name='x0', val=np.ones(nn), desc='distance the arm protrudes from the pivot', units='m')
        self.add_input(name='x1', val=np.ones(nn), desc='horizontal angle of the arm', units='rad')
        self.add_input(name='x2', val=np.ones(nn), desc='vertical angle of the arm', units='rad')
        self.add_input(name='x3', val=np.ones(nn), desc='rate of change of arm protrusion', units='m/s')
        self.add_input(name='x4', val=np.ones(nn), desc='horizontal angular velocity', units='rad/s')
        self.add_input(name='x5', val=np.ones(nn), desc='vertical angular velocity', units='rad/s')

        self.add_input(name='u0', val=np.ones(nn), desc='arm length control', units='m**2/s**2')
        self.add_input(name='u1', val=np.ones(nn), desc='horizontal angle control', units='m**3*rad/s**2')
        self.add_input(name='u2', val=np.ones(nn), desc='vertical angle control', units='m**3*rad/s**2')

        # outputs: 6 state rates
        self.add_output(name='x0_dot', val=np.ones(nn), desc='rate of change of arm protrusion', units='m/s')
        self.add_output(name='x1_dot', val=np.ones(nn), desc='horizontal angular velocity', units='rad/s')
        self.add_output(name='x2_dot', val=np.ones(nn), desc='vertical angular velocity', units='rad/s')
        self.add_output(name='x3_dot', val=np.ones(nn), desc='second derivative of arm protrusion', units='m/s**2')
        self.add_output(name='x4_dot', val=np.ones(nn), desc='horizontal angular acceleration', units='rad/s**2')
        self.add_output(name='x5_dot', val=np.ones(nn), desc='vertical angular acceleration', units='rad/s**2')

        # set up the partials
        r = c = np.arange(nn)

        self.declare_partials(of='x0_dot', wrt='x3', rows=r, cols=c, val=1.0)

        self.declare_partials(of='x1_dot', wrt='x4', rows=r, cols=c, val=1.0)

        self.declare_partials(of='x2_dot', wrt='x5', rows=r, cols=c, val=1.0)

        self.declare_partials(of='x3_dot', wrt='x3', rows=r, cols=c, val=1.0)
        self.declare_partials(of='x3_dot', wrt='u0', rows=r, cols=c, val=1.0/L)

        self.declare_partials(of='x4_dot', wrt='x0', rows=r, cols=c)
        self.declare_partials(of='x4_dot', wrt='x2', rows=r, cols=c)
        self.declare_partials(of='x4_dot', wrt='x4', rows=r, cols=c, val=1.0)
        self.declare_partials(of='x4_dot', wrt='u1', rows=r, cols=c)

        self.declare_partials(of='x5_dot', wrt='x0', rows=r, cols=c)
        self.declare_partials(of='x5_dot', wrt='x5', rows=r, cols=c, val=1.0)
        self.declare_partials(of='x5_dot', wrt='u2', rows=r, cols=c)

    def compute(self, inputs, outputs):
        L = self.options['arm_length']

        x0 = inputs['x0']
        # x1 = inputs['x1']
        x2 = inputs['x2']
        x3 = inputs['x3']
        x4 = inputs['x4']
        x5 = inputs['x5']

        u0 = inputs['u0']
        u1 = inputs['u1']
        u2 = inputs['u2']

        outputs['x0_dot'] = x3
        outputs['x1_dot'] = x4
        outputs['x2_dot'] = x5
        outputs['x3_dot'] = u0/L
        outputs['x4_dot'] = 3*u1/(((L - x0)**3 + x0**3)*np.sin(x2)**2)
        outputs['x5_dot'] = 3*u2/((L - x0)**3 + x0**3)

    def compute_partials(self, inputs, partials):
        L = self.options['arm_length']

        x0 = inputs['x0']
        x2 = inputs['x2']

        u1 = inputs['u1']
        u2 = inputs['u2']

        partials['x4_dot', 'x0'] = 9*u1*(L - 2*x0)/(L*(L**2 - 3*L*x0 + 3*x0**2)**2*np.sin(x2)**2)
        partials['x4_dot', 'x2'] = -24*u1*np.cos(x2)/(((L - x0)**3 + x0**3)*(3*np.sin(x2) - np.sin(3*x2)))
        partials['x4_dot', 'u1'] = 3/(((L - x0)**3 + x0**3)*np.sin(x2)**2)

        partials['x5_dot', 'x0'] = 9*u2*(L - 2*x0)/(L*(L**2 - 3*L*x0 + 3*x0**2)**2)
        partials['x5_dot', 'u2'] = 3/((L - x0)**3 + x0**3)



import openmdao.api as om
from openmdao.utils.testing_utils import require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


@require_pyoptsparse(optimizer='SLSQP')
def brachistochrone_min_time(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                             compressed=True, optimizer='SLSQP', run_driver=True, force_alloc_complex=False,
                             solve_segments=False):
    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    if optimizer == 'SNOPT':
        # p.driver.opt_settings['iSumm'] = 6
        p.driver.opt_settings['Verify level'] = 3
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['print_level'] = 0
    p.driver.declare_coloring(tol=1.0E-12)

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
    elif transcription == 'radau-ps':
        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)
    elif transcription == 'shooting-gauss-lobatto':
        grid = dm.GaussLobattoGrid(num_segments=num_segments,
                                   nodes_per_seg=transcription_order,
                                   compressed=compressed)
        t = dm.ExplicitShooting(grid=grid)
    elif transcription == 'shooting-radau':
        grid = dm.RadauGrid(num_segments=num_segments,
                            nodes_per_seg=transcription_order + 1,
                            compressed=compressed)
        t = dm.ExplicitShooting(grid=grid)
    elif transcription == 'birkhoff':
        t = dm.Birkhoff(num_nodes=transcription_order + 1,
                        grid_type='cgl')
        # phase = dm.ImplicitPhase(ode_class=BrachistochroneODE, num_nodes=11)

    traj = dm.Trajectory()
    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
    p.model.add_subsystem('traj0', traj)
    traj.add_phase('phase0', phase)

    phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10))

    phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=solve_segments)
    phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=solve_segments)

    # Note that by omitting the targets here Dymos will automatically attempt to connect
    # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
    phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=solve_segments)

    phase.add_control('theta',
                      continuity=True, rate_continuity=True,
                      units='deg', lower=0.01, upper=179.9)

    phase.add_parameter('g', targets=['g'], units='m/s**2')

    phase.add_timeseries('timeseries2',
                         transcription=dm.Radau(num_segments=num_segments*5,
                                                order=transcription_order,
                                                compressed=compressed),
                         subset='control_input')

    phase.add_boundary_constraint('x', loc='final', equals=10)
    phase.add_boundary_constraint('y', loc='final', equals=5)
    phase.add_path_constraint('y', lower=0, upper=20)
    # Minimize time at the end of the phase
    phase.add_objective('time_phase', loc='final', scaler=10)

    phase.add_timeseries_output('check')

    p.setup(check=['unconnected_inputs'], force_alloc_complex=force_alloc_complex)

    phase.set_simulate_options(method='RK23')

    phase.set_time_val(initial=0.0, duration=2.0)

    phase.set_state_val('x', [0, 10])
    phase.set_state_val('y', [10, 5])
    phase.set_state_val('v', [0, 9.9])

    phase.set_control_val('theta', [5, 100])
    phase.set_parameter_val('g', 9.80665)

    dm.run_problem(p, run_driver=run_driver, simulate=True, make_plots=True,
                   simulate_kwargs={'times_per_seg': 100})

    return p


if __name__ == '__main__':

    with dm.options.temporary(include_check_partials=True):
        p = brachistochrone_min_time(transcription='birkhoff', num_segments=1, run_driver=True,
                                     transcription_order=19, compressed=False, optimizer='SLSQP',
                                     solve_segments=False, force_alloc_complex=True)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_near_equal

from dymos.utils.misc import om_version


class BrachistochroneODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')
        self.add_input('g', val=9.80665, desc='acceleration of gravity', units='m/s**2')
        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')
        self.add_input('final_time', val=0.0, desc='expected final time', units='s')
        self.add_output('xdot', val=np.zeros(nn), desc='horizontal velocity', units='m/s')
        self.add_output('ydot', val=np.zeros(nn), desc='vertical velocity', units='m/s')
        self.add_output('vdot', val=np.zeros(nn), desc='acceleration mag.', units='m/s**2')

        # Setup partials
        arange = np.arange(self.options['num_nodes'], dtype=int)

        self.declare_partials(of='vdot', wrt='g', rows=arange, cols=np.zeros(nn, dtype=int))
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='xdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='ydot', wrt='theta', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        jacobian['vdot', 'g'] = cos_theta
        jacobian['vdot', 'theta'] = -g * sin_theta

        jacobian['xdot', 'v'] = sin_theta
        jacobian['xdot', 'theta'] = v * cos_theta

        jacobian['ydot', 'v'] = -cos_theta
        jacobian['ydot', 'theta'] = v * sin_theta


@use_tempdirs
class TestBrachistochroneTimeseriesFeedback(unittest.TestCase):

    def test_timeseries_feedback(self):
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        phase.add_parameter('final_time', units='s', static_target=True, opt=False)
        traj.add_parameter('final_time', units='s', static_target=True, opt=False)

        traj.connect(src_name='phase0.timeseries.time', tgt_name='parameters:final_time', src_indices=om.slicer[-1, ...])

        # The feedback connection introduced iterative behavior, so now nonlinear solver and linear solver are needed.
        traj.nonlinear_solver = om.NonlinearBlockGS()
        traj.linear_solver = om.DirectSolver()

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that the it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, rate_source='ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, rate_source='vdot')

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi)

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 5], units='m/s')

        phase.set_control_val('theta', [90, 90])
        phase.set_parameter_val('final_time', 0.0, units='s')

        # Run the driver to solve the problem
        dm.run_problem(p, run_driver=True, simulate=True)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol_case = om.CaseReader(sol_db).get_case('final')
        sim_case = om.CaseReader(sim_db).get_case('final')

        final_time_timeseries_sol = sol_case.get_val('traj.phase0.timeseries.time')[-1, ...]
        final_time_traj_param_sol = sol_case.get_val('traj.parameter_vals:final_time')[0, 0]
        final_time_timeseries_sim = sim_case.get_val('traj.phase0.timeseries.time')[-1, ...]
        final_time_traj_param_sim = sim_case.get_val('traj.parameter_vals:final_time')[0, 0]

        p.model.list_outputs(prom_name=True)

        assert_near_equal(final_time_traj_param_sol, final_time_timeseries_sol)
        assert_near_equal(final_time_traj_param_sim, final_time_timeseries_sol)
        assert_near_equal(final_time_timeseries_sim, final_time_timeseries_sol)




import numpy as np

import openmdao
import openmdao.api as om
from openmdao.core.constants import _ReprClass

from .constants import INF_BOUND
from .indexing import get_desvar_indices


# unique object to check if default is given (when None is an allowed value)
_unspecified = _ReprClass("unspecified")
_none_or_unspecified = {None, _unspecified}


def get_rate_units(units, time_units, deriv=1):
    """
    Return a string for rate units given units for the variable and time units.

    Parameters
    ----------
    units : str
        Units of a given variable.
    time_units : str
        Time units.
    deriv : int
        If 1, provide the units of the first derivative.  If 2,
        provide the units of the second derivative.

    Returns
    -------
    str
        Corresponding rate units for the given variable.
    """
    if deriv not in (1, 2):
        raise ValueError('deriv argument must be 1 or 2.')

    tu = time_units if deriv == 1 else f'{time_units}**2'

    if units is not None and time_units is not None:
        return f'{units}/{tu}'
    elif units is not None:
        return units
    elif time_units is not None:
        return f'1.0/{tu}'


def reshape_val(val, shape, num_input_nodes):
    """
    Return the given value reshaped to (num_input_nodes,) + shape.

    If the value is scalar or a size-1 array, return that value multiplied by
    np.ones((num_input_nodes,) + shape).  If the value's shape is shape, then
    repeat those values along a new first dimension.  Otherwise, reshape it to
    the correct shape and return that.

    Parameters
    ----------
    val : float or array-like
        The values to be conformed to the desired shape.
    shape : tuple
        The shape of the desired output at each node.
    num_input_nodes : int
        The number of nodes along which the value is repeated.

    Returns
    -------
    np.array
        The given value of the correct shape.
    """
    if np.isscalar(val) or np.prod(np.asarray(val).shape) == 1:
        shaped_val = float(val) * np.ones((num_input_nodes,) + shape)
    elif np.asarray(val).shape == shape:
        shaped_val = np.repeat(val[np.newaxis, ...], num_input_nodes, axis=0)
    else:
        shaped_val = np.reshape(val, (num_input_nodes,) + shape)
    return shaped_val


def _format_phase_constraint_alias(phase, con_name, con_type, indices=None):
    """
    Get an alias for a constraint of the given type on the given path in the given phase.

    Parameters
    ----------
    phase : Phase
        The dymos phase to which the constraint belongs.
    con_name : str
        The name or path of the constraint.
    con_type : str
        One of 'initial', 'final', or 'path'.
    indices : tuple or None
        The indices of the constraint variable to be constrained. These indices
        ignore the num_nodes dimension in path constraints.

    Returns
    -------
    str
        The alias of the constraint.
    """
    str_idxs = '' if indices is None else f'[{indices}]'
    if f'.phases.{phase.name}' in phase.pathname:
        phase_path = phase.pathname.replace(f'.phases.{phase.name}', f'.{phase.name}')
    else:
        phase_path = phase.pathname
    # return f'{phase_path}->{con_type}->{con_name}{str_idxs}'
    return f'{phase_path}.{con_name}[{con_type}]{str_idxs}'


class CoerceDesvar(object):
    """
    Check the desvar options for the appropriate shape and resize accordingly with options.

    Parameters
    ----------
    num_input_nodes : int
        Number of input nodes.
    desvar_indices : ndarray
        Flattened indices of the variable.
    options : dict
        Variable options dictionary, should contain "shape".
    """
    def __init__(self, num_input_nodes, desvar_indices=None, options=None):
        self.num_input_nodes = num_input_nodes
        shape = options['shape']
        size = np.prod(shape, dtype=int)

        if desvar_indices is None:
            desvar_indices = get_desvar_indices(size, num_input_nodes,
                                                options['fix_initial'], options['fix_final'])
        self.desvar_indices = desvar_indices
        self.options = options

    def __call__(self, option):
        """
        Test that an option's shape is compliant with the number of input nodes for the design variable.

        Parameters
        ----------
        option : str
            The name of the option whose value(s) are desired.

        Returns
        -------
        object
            The value of the desvar option.

        Raises
        ------
        ValueError
            If the number of values in the option is not compliant with the number of input
            nodes for the design variable.

        """
        val = self.options[option]
        if option == 'lower':
            lb = np.zeros_like(self.desvar_indices, dtype=float)
            lb[:] = -INF_BOUND if self.options['lower'] is None else val
            return lb

        if option == 'upper':
            ub = np.zeros_like(self.desvar_indices, dtype=float)
            ub[:] = INF_BOUND if self.options['upper'] is None else val
            return ub

        if val is None or np.isscalar(val):
            return val
        # Handle value for vector/matrix valued variables
        if isinstance(val, list):
            val = np.asarray(val)
        if val.shape == self.options['shape']:
            return np.tile(val.flatten(), int(len(self.desvar_indices)/val.size))
        else:
            raise ValueError('array-valued option {0} must have same shape '
                             'as states ({1})'.format(option, self.options['shape']))


def CompWrapperConfig(comp_class, config_io_args=None):
    """
    Returns a wrapped comp_class that calls its configure_io method at the end of setup.

    This allows for standalone testing of Dymos components that normally require their parent group
    to configure them.

    Parameters
    ----------
    comp_class : Component class
        Class that we would like to wrap.
    config_io_args : list
        Arguments to be passed to config_io.

    Returns
    -------
    WrappedClass
        Wrapped version of comp_class.
    """
    class WrappedClass(comp_class):

        def setup(self):
            """
            Appends a call to configure_io after setup.
            """
            super(WrappedClass, self).setup()
            args = [] if config_io_args is None else config_io_args
            self.configure_io(*args)

    return WrappedClass


# Modify class so we can run it standalone.
def GroupWrapperConfig(comp_class, config_io_args=None):
    """
    Returns a wrapped group_class that calls its configure_io method at the end of setup.

    This allows for standalone testing of Dymos components that normally require their parent group
    to configure them.

    Parameters
    ----------
    comp_class : Group class
       Class that we would like to wrap.
    config_io_args : list
        Arguments to be passed to config_io.

    Returns
    -------
    WrappedClass
        Wrapped version of comp_class.
    """
    class WrappedClass(comp_class):

        def setup(self):
            """
            Setup as normal.
            """
            super(WrappedClass, self).setup()

        def configure(self):
            """
            Call configure_io during configure.
            """
            args = [] if config_io_args is None else config_io_args
            self.configure_io(*args)

    return WrappedClass


def create_subprob(base_name, comm, reports=False):
    """
    Create a new problem using basename possibly appended with unique identifiers if name collisions occur.

    Parameters
    ----------
    base_name : str
        The base name of the problem. This may be appended by `_{int}` to obtain a unique problem name.
        In the event of running under MPI, an 8-character hash may further append the name to ensure
        it is unique.
    comm : comm
        The MPI comm to be used by the subproblem.
    reports : bool or None or str or Sequence
        Reports setting for the subproblems run under simualate.

    Returns
    -------
    Problem
        The instantiated OpenMDAO problem instance.
    """
    from openmdao.core.problem import _problem_names

    # Find a unique sim problem name. This mostly causes problems
    # when many simulations are being run in a single process, as in testing.
    i = 0
    sim_prob_name = f'{base_name}_{i}'
    while sim_prob_name in _problem_names:
        i += 1
        sim_prob_name = f'{base_name}_{i}'

    try:
        p = om.Problem(comm=comm, reports=reports, name=sim_prob_name)
    except ValueError:
        # Testing under MPI, we still might have name collisions. In that case, add a random hash
        # to the end of the problem name.
        import hashlib
        str_hash = hashlib.sha256(used_for_security=False)[:8]
        p = om.Problem(comm=comm, reports=reports, name=f'{sim_prob_name}_{str_hash}')
    return p


def om_version():
    """
    Return version infromation for OpenMDAO.

    This information is useful for executing code that requires a specific
    version of OpenMDAO. The tuple format returned by this function can
    easily be compared using a statement like `if om_version()[0] < (3, 3, 0)`.

    Returns
    -------
    tuple
        The semantic version of OpenMDAO in a comparable tuple.
    str
        One of "dev" or "release".
    """
    try:
        numeric, rel = openmdao.__version__.split('-')
    except ValueError:
        numeric = openmdao.__version__
        rel = 'release'
    return tuple([int(s) for s in numeric.split('.')]), rel



import numpy as np
import openmdao.api as om
import openmdao.utils.units as units

# Add canonical units to OpenMDAO
MU_earth = 3.986592936294783e14
R_earth = 6378137.0

period = 2 * np.pi * np.sqrt(R_earth**3 / MU_earth)

# Add canonical time and distance units for these EOM
units.add_unit('TU', '{0}*s'.format(period))
units.add_unit('DU', '{0}*m'.format(R_earth))


class FiniteBurnODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('r',
                       val=np.ones(nn),
                       desc='radius from center of attraction',
                       units='DU')

        self.add_input('theta',
                       val=np.zeros(nn),
                       desc='anomaly term',
                       units='rad')

        self.add_input('vr',
                       val=np.zeros(nn),
                       desc='local vertical velocity component',
                       units='DU/TU')

        self.add_input('vt',
                       val=np.zeros(nn),
                       desc='local horizontal velocity component',
                       units='DU/TU')

        self.add_input('accel',
                       val=np.zeros(nn),
                       desc='acceleration due to thrust',
                       units='DU/TU**2')

        self.add_input('u1',
                       val=np.zeros(nn),
                       desc='thrust angle above local horizontal',
                       units='rad')

        self.add_input('c',
                       val=np.zeros(nn),
                       desc='exhaust velocity',
                       units='DU/TU')

        self.add_output('r_dot',
                        val=np.ones(nn),
                        desc='rate of change of radius from center of attraction',
                        units='DU/TU')

        self.add_output('theta_dot',
                        val=np.zeros(nn),
                        desc='rate of change of anomaly term',
                        units='rad/TU')

        self.add_output('vr_dot',
                        val=np.zeros(nn),
                        desc='rate of change of local vertical velocity component',
                        units='DU/TU**2')

        self.add_output('vt_dot',
                        val=np.zeros(nn),
                        desc='rate of change of local horizontal velocity component',
                        units='DU/TU**2')

        self.add_output('at_dot',
                        val=np.zeros(nn),
                        desc='rate of change of acceleration due to thrust',
                        units='DU/TU**3')

        self.add_output('deltav_dot',
                        val=np.zeros(nn),
                        desc='rate of change of delta-V',
                        units='DU/TU**2')

        self.add_output('pos_x',
                        val=np.zeros(nn),
                        desc='x-component of position',
                        units='DU')

        self.add_output('pos_y',
                        val=np.zeros(nn),
                        desc='x-component of position',
                        units='DU')

        # Setup partials
        ar = np.arange(self.options['num_nodes'])

        # r dot is a linear function of vr, so provide the partial value here
        self.declare_partials(of='r_dot', wrt='vr', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='theta_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='theta_dot', wrt='vt', rows=ar, cols=ar)

        self.declare_partials(of='vr_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='vr_dot', wrt='vt', rows=ar, cols=ar)
        self.declare_partials(of='vr_dot', wrt='accel', rows=ar, cols=ar)
        self.declare_partials(of='vr_dot', wrt='u1', rows=ar, cols=ar)

        self.declare_partials(of='vt_dot', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='vr', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='vt', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='accel', rows=ar, cols=ar)
        self.declare_partials(of='vt_dot', wrt='u1', rows=ar, cols=ar)

        self.declare_partials(of='at_dot', wrt='accel', rows=ar, cols=ar)
        self.declare_partials(of='at_dot', wrt='c', rows=ar, cols=ar)

        self.declare_partials(of='deltav_dot', wrt='accel', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='pos_x', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='pos_x', wrt='theta', rows=ar, cols=ar)

        self.declare_partials(of='pos_y', wrt='r', rows=ar, cols=ar)
        self.declare_partials(of='pos_y', wrt='theta', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        r = inputs['r']
        theta = inputs['theta']
        vr = inputs['vr']
        vt = inputs['vt']
        at = inputs['accel']
        u1 = inputs['u1']
        c = inputs['c']

        outputs['r_dot'] = vr
        outputs['theta_dot'] = vt / r
        outputs['vr_dot'] = vt**2 / r - 1 / r**2 + at * np.sin(u1)
        outputs['vt_dot'] = -vr * vt / r + at * np.cos(u1)
        outputs['at_dot'] = at**2 / c
        outputs['deltav_dot'] = at

        outputs['pos_x'] = r * np.cos(theta)
        outputs['pos_y'] = r * np.sin(theta)

    def compute_partials(self, inputs, partials):
        r = inputs['r']
        theta = inputs['theta']
        vr = inputs['vr']
        vt = inputs['vt']
        at = inputs['accel']
        u1 = inputs['u1']
        c = inputs['c']

        su1 = np.sin(u1)
        cu1 = np.cos(u1)

        partials['theta_dot', 'r'] = -vt / r**2
        partials['theta_dot', 'vt'] = 1.0 / r

        partials['vr_dot', 'r'] = -vt**2 / r**2 + 2.0 / r**3
        partials['vr_dot', 'vt'] = 2 * vt / r
        partials['vr_dot', 'accel'] = su1
        partials['vr_dot', 'u1'] = at * cu1

        partials['vt_dot', 'r'] = vr * vt / r**2
        partials['vt_dot', 'vr'] = -vt / r
        partials['vt_dot', 'vt'] = -vr / r
        partials['vt_dot', 'accel'] = cu1
        partials['vt_dot', 'u1'] = -at * su1

        partials['at_dot', 'accel'] = 2 * at / c
        partials['at_dot', 'c'] = -at**2 / c**2

        partials['pos_x', 'r'] = np.cos(theta)
        partials['pos_x', 'theta'] = -r * np.sin(theta)

        partials['pos_y', 'r'] = np.sin(theta)
        partials['pos_y', 'theta'] = r * np.cos(theta)



import numpy as np

from ..grid_data import GaussLobattoGrid
from ...utils.misc import get_rate_units

from ..common.continuity_comp import ContinuityCompBase


class ExplicitShootingContinuityComp(ContinuityCompBase):
    """
    ContinuityComp defines constraints to ensure continuity between adjacent segments.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._controls_to_enforce = set()
        self._control_rates_to_enforce = set()
        self._control_rates2_to_enforce = set()

        self.rate_jac_templates = {}
        self.name_maps = {}

    def _configure_state_continuity(self):
        # TODO This method will be used when multiple shooting is implemented.
        pass

        # state_options = self.options['state_options']
        # num_segments = self.options['grid_data'].num_segments
        # compressed = self.options['grid_data'].compressed
        #
        # if num_segments <= 1:
        #     return
        #
        # super(ExplicitShootingContinuityComp, self)._configure_state_continuity()
        #
        # for state_name, options in state_options.items():
        #     if options['continuity'] and not compressed:
        #         # State continuity is nonlinear in (TBD) explicit multiple shooting phases
        #         self.add_constraint(name=f'defect_states:{state_name}',
        #                             equals=0.0, scaler=1.0, linear=False)

    def _configure_control_continuity(self, controls_to_enforce=None, control_rates_to_enforce=None,
                                      control_rates2_to_enforce=None):
        """
        Configures control continuity.

        Each argument contains the names of those variables which require continuity.

        Parameters
        ----------
        controls : set or Sequence of str or None
            The names of the controls whose values are to be enforced at segment boundaries.
        control_rates : set or Sequence of str or None
            The names of controls whose rates are to be enforced at segment boundaries.
        control_rates2 : set or Sequence of str or None
            The names of controls whose second derivatives are to be enforced at the segment boundaries.
        """

        control_options = self.options['control_options']
        num_segend_nodes = self.options['grid_data'].subset_num_nodes['segment_ends']
        num_segments = self.options['grid_data'].num_segments
        time_units = self.options['time_units']

        self._controls_to_enforce = controls_to_enforce
        self._control_rates_to_enforce = control_rates_to_enforce
        self._control_rates2_to_enforce = control_rates2_to_enforce

        if num_segments <= 1:
            # Control value and rate continuity is enforced even with compressed transcription
            return

        any_rate_continuity = False

        for control_name, options in control_options.items():
            shape = options['shape']
            size = np.prod(shape)
            units = options['units']
            rate_units = get_rate_units(units, time_units, deriv=1)
            rate2_units = get_rate_units(units, time_units, deriv=2)

            # Define the sparsity pattern for rate and rate2 continuity
            rs_size1 = np.repeat(np.arange(num_segments - 1, dtype=int), 2)
            cs_size1 = np.arange(1, num_segend_nodes - 1, dtype=int)

            template = np.zeros((num_segments - 1, num_segend_nodes))
            template[rs_size1, cs_size1] = 1.0
            template = np.kron(template, np.eye(size))
            rs, cs = template.nonzero()

            vals = np.zeros(len(rs), dtype=float)
            vals[0::2] = -1.0
            vals[1::2] = 1.0
            self.rate_jac_templates[control_name] = vals

            #
            # Setup value continuity
            #
            self.name_maps[control_name] = {}

            if control_name in controls_to_enforce:
                self.name_maps[control_name]['value_names'] = \
                    (f'controls:{control_name}',
                     f'defect_controls:{control_name}')

                self.add_input(
                    name=f'controls:{control_name}',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} at segment endpoint nodes',
                    units=units)

                self.add_output(
                    name=f'defect_controls:{control_name}',
                    val=5*np.ones((num_segments - 1,) + shape),
                    desc=f'Continuity constraint values for control {control_name}',
                    units=units)

                self.declare_partials(
                    f'defect_controls:{control_name}',
                    f'controls:{control_name}',
                    val=vals, rows=rs, cols=cs,
                )

                linear_cnty = isinstance(self.options['grid_data'], GaussLobattoGrid)
                self.add_constraint(name=f'defect_controls:{control_name}',
                                    equals=0.0, scaler=1.0, linear=linear_cnty)

            if control_name in control_rates_to_enforce:
                any_rate_continuity = True

                self.name_maps[control_name]['rate_names'] = \
                    (f'control_rates:{control_name}_rate',
                     f'defect_control_rates:{control_name}_rate')

                self.add_input(
                    name=f'control_rates:{control_name}_rate',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} derivative at segment endpoint nodes',
                    units=rate_units)

                self.add_output(
                    name=f'defect_control_rates:{control_name}_rate',
                    shape=(num_segments - 1,) + shape,
                    desc=f'Consistency constraint values for control {control_name} derivative',
                    units=rate_units)

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate',
                    f'control_rates:{control_name}_rate',
                    rows=rs, cols=cs,
                )

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate',
                    't_duration', dependent=True,
                )

                self.add_constraint(name=f'defect_control_rates:{control_name}_rate',
                                    equals=0.0, scaler=options['rate_continuity_scaler'],
                                    linear=False)

            if control_name in control_rates2_to_enforce:
                any_rate_continuity = True

                self.name_maps[control_name]['rate2_names'] = \
                    (f'control_rates:{control_name}_rate2',
                     f'defect_control_rates:{control_name}_rate2')

                self.add_input(
                    name=f'control_rates:{control_name}_rate2',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} second derivative '
                         'at discretization nodes',
                    units=rate2_units)

                self.add_output(
                    name=f'defect_control_rates:{control_name}_rate2',
                    shape=(num_segments - 1,) + shape,
                    desc='Consistency constraint values for control '
                         f'{control_name} second derivative',
                    units=rate2_units)

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate2',
                    f'control_rates:{control_name}_rate2',
                    rows=rs, cols=cs,
                )

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate2',
                    't_duration', dependent=True
                )

                self.add_constraint(name=f'defect_control_rates:{control_name}_rate2',
                                    equals=0.0, scaler=options['rate2_continuity_scaler'],
                                    linear=False)

        if any_rate_continuity:
            self.add_input('t_duration', units=time_units, val=1.0, desc='time duration of the phase')

    def configure_io(self, controls_to_enforce=None, control_rates_to_enforce=None,
                     control_rates2_to_enforce=None):
        """
        Configures control continuity.

        Each argument contains the names of those variables which require continuity.

        Parameters
        ----------
        controls_to_enforce : set or Sequence of str or None
            The names of the controls whose values are to be enforced at segment boundaries.
        control_rates_to_enforce : set or Sequence of str or None
            The names of controls whose rates are to be enforced at segment boundaries.
        control_rates2_to_enforce : set or Sequence of str or None
            The names of controls whose second derivatives are to be enforced at the segment boundaries.
        """
        self.rate_jac_templates = {}
        self.name_maps = {}

        self._configure_control_continuity(controls_to_enforce=controls_to_enforce,
                                           control_rates_to_enforce=control_rates_to_enforce,
                                           control_rates2_to_enforce=control_rates2_to_enforce)
        self._configure_state_continuity()

    def _compute_state_continuity(self, inputs, outputs):
        pass

    def _compute_control_continuity(self, inputs, outputs):
        control_options = self.options['control_options']

        if any([options['rate_continuity'] or options['rate2_continuity'] for options in control_options.values()]):
            dt_dptau = inputs['t_duration'] / 2.0

        for name, options in control_options.items():
            if name in self._controls_to_enforce:
                input_name, output_name = self.name_maps[name]['value_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = start_vals - end_vals

            if name in self._control_rates_to_enforce:
                input_name, output_name = self.name_maps[name]['rate_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = (start_vals - end_vals) * dt_dptau

            if name in self._control_rates2_to_enforce:
                input_name, output_name = self.name_maps[name]['rate2_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = (start_vals - end_vals) * dt_dptau ** 2



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.trajectory.phase_linkage_comp import PhaseLinkageComp
from dymos.trajectory.options import LinkageOptionsDictionary
from dymos.utils.testing_utils import assert_check_partials


@use_tempdirs
class TestPhaseLinkageComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True
        self.p = om.Problem(model=om.Group())

        ivp = self.p.model.add_subsystem('ivc', subsys=om.IndepVarComp(), promotes_outputs=['*'])

        ndn = 20
        nn = 30

        ivp.add_output('phase0:x', val=np.zeros((ndn, 1)), units='m')
        ivp.add_output('phase0:u', val=np.zeros((nn, 3)), units='deg')
        ivp.add_output('phase0:v', val=np.zeros((nn, 3, 3)), units='N')

        ivp.add_output('phase1:x', val=np.zeros((ndn, 1)), units='m')
        ivp.add_output('phase1:u', val=np.zeros((nn, 3)), units='deg')
        ivp.add_output('phase1:v', val=np.zeros((nn, 3, 3)), units='N')

        linkage_comp = PhaseLinkageComp()

        x_lnk = LinkageOptionsDictionary()
        x_lnk['phase_a'] = 'phase0'
        x_lnk['phase_b'] = 'phase1'
        x_lnk['var_a'] = 'x'
        x_lnk['var_b'] = 'x'
        x_lnk['loc_a'] = 'final'
        x_lnk['loc_b'] = 'initial'
        x_lnk['units'] = 'm'
        x_lnk['shape'] = (1,)
        x_lnk['equals'] = 0.0

        u_lnk = LinkageOptionsDictionary()
        u_lnk['phase_a'] = 'phase0'
        u_lnk['phase_b'] = 'phase1'
        u_lnk['var_a'] = 'u'
        u_lnk['var_b'] = 'u'
        u_lnk['loc_a'] = 'final'
        u_lnk['loc_b'] = 'initial'
        u_lnk['units'] = 'deg'
        u_lnk['shape'] = (3,)
        u_lnk['equals'] = 0.0

        v_lnk = LinkageOptionsDictionary()
        v_lnk['phase_a'] = 'phase0'
        v_lnk['phase_b'] = 'phase1'
        v_lnk['var_a'] = 'v'
        v_lnk['var_b'] = 'v'
        v_lnk['loc_a'] = 'final'
        v_lnk['loc_b'] = 'initial'
        v_lnk['units'] = 'N'
        v_lnk['shape'] = (3, 3)
        v_lnk['equals'] = 0.0

        linkage_comp.add_linkage_configure(x_lnk)
        linkage_comp.add_linkage_configure(u_lnk)
        linkage_comp.add_linkage_configure(v_lnk)

        self.p.model.add_subsystem('linkage_comp', subsys=linkage_comp)

        self.p.model.connect('phase0:x', 'linkage_comp.phase0:x',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.model.connect('phase1:x', 'linkage_comp.phase1:x',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.model.connect('phase0:u', 'linkage_comp.phase0:u',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.model.connect('phase1:u', 'linkage_comp.phase1:u',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.model.connect('phase0:v', 'linkage_comp.phase0:v',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.model.connect('phase1:v', 'linkage_comp.phase1:v',
                             src_indices=om.slicer[[0, -1], ...])

        self.p.setup()

        self.p['phase0:x'] = np.random.rand(*self.p['phase0:x'].shape)
        self.p['phase0:u'] = np.random.rand(*self.p['phase0:u'].shape)
        self.p['phase0:v'] = np.random.rand(*self.p['phase0:v'].shape)

        self.p['phase1:x'] = np.random.rand(*self.p['phase1:x'].shape)
        self.p['phase1:u'] = np.random.rand(*self.p['phase1:u'].shape)
        self.p['phase1:v'] = np.random.rand(*self.p['phase1:v'].shape)

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):

        assert_almost_equal(-self.p['phase1:x'][0, ...] + self.p['phase0:x'][-1, ...],
                            self.p['linkage_comp.phase0:x_final|phase1:x_initial'])

        assert_almost_equal(-self.p['phase1:u'][0, ...] + self.p['phase0:u'][-1, ...],
                            self.p['linkage_comp.phase0:u_final|phase1:u_initial'])

        assert_almost_equal(-self.p['phase1:v'][0, ...] + self.p['phase0:v'][-1, ...],
                            self.p['linkage_comp.phase0:v_final|phase1:v_initial'])

    def test_partials(self):
        cpd = self.p.check_partials(out_stream=None)
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om
from ...models.atmosphere import USatm1976Comp
from .aero import AeroGroup
from .prop import PropGroup
from ...models.eom import FlightPathEOM2D


class MinTimeClimbODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='atmos',
                           subsys=USatm1976Comp(num_nodes=nn, h_def='geodetic'),
                           promotes_inputs=['h'])

        self.add_subsystem(name='aero',
                           subsys=AeroGroup(num_nodes=nn),
                           promotes_inputs=['v', 'alpha', 'S'])

        self.connect('atmos.sos', 'aero.sos')
        self.connect('atmos.rho', 'aero.rho')

        self.add_subsystem(name='prop',
                           subsys=PropGroup(num_nodes=nn),
                           promotes_inputs=['h', 'Isp', 'throttle'])

        self.connect('aero.mach', 'prop.mach')

        self.add_subsystem(name='flight_dynamics',
                           subsys=FlightPathEOM2D(num_nodes=nn),
                           promotes_inputs=['m', 'v', 'gam', 'alpha'])

        self.connect('aero.f_drag', 'flight_dynamics.D')
        self.connect('aero.f_lift', 'flight_dynamics.L')
        self.connect('prop.thrust', 'flight_dynamics.T')

        self.set_input_defaults('h', val=1.0)
        self.set_input_defaults('alpha', val=0.0)



import numpy as np

import openmdao.api as om


class MachComp(om.ExplicitComponent):
    """ Compute the Mach number based on vehicle airspeed and local speed
    of sound.
    """
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input(name='v', shape=(nn,), desc='velocity magnitude', units='m/s')
        self.add_input('sos', shape=(nn,), desc='alpha lift coefficient', units='m/s')

        # Outputs
        self.add_output(name='mach', val=0.7*np.ones(nn), desc='Mach number', units=None)

        # Jacobian
        ar = np.arange(nn)
        self.declare_partials(of='mach', wrt='v', rows=ar, cols=ar)
        self.declare_partials(of='mach', wrt='sos', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        outputs['mach'][:] = inputs['v'] / inputs['sos']

    def compute_partials(self, inputs, partials):
        partials['mach', 'v'] = 1.0 / inputs['sos']
        partials['mach', 'sos'] = -inputs['v'] / inputs['sos'] ** 2



import unittest

import numpy as np

import openmdao.api as om

import dymos
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.misc import GroupWrapperConfig
from dymos.utils.testing_utils import PhaseStub, SimpleODE
from dymos.transcriptions.pseudospectral.components import BirkhoffIterGroup
from dymos.phase.options import StateOptionsDictionary, TimeOptionsDictionary
from dymos.transcriptions.grid_data import BirkhoffGrid


BirkhoffIterGroup = GroupWrapperConfig(BirkhoffIterGroup, [PhaseStub()])


@use_tempdirs
class TestBirkhoffIterGroup(unittest.TestCase):

    def test_solve_segments_gl_fwd(self):
        for grid_type in ['lgl', 'cgl']:
            with self.subTest(msg=grid_type):
                with dymos.options.temporary(include_check_partials=True):

                    state_options = {'x': StateOptionsDictionary()}

                    state_options['x']['shape'] = (1,)
                    state_options['x']['units'] = 's**2'
                    state_options['x']['targets'] = ['x']
                    state_options['x']['initial_bounds'] = (None, None)
                    state_options['x']['final_bounds'] = (None, None)
                    state_options['x']['solve_segments'] = 'forward'
                    state_options['x']['rate_source'] = 'x_dot'

                    time_options = TimeOptionsDictionary()
                    grid_data = BirkhoffGrid(num_nodes=31, grid_type=grid_type)
                    nn = grid_data.subset_num_nodes['all']
                    ode_class = SimpleODE

                    p = om.Problem()
                    p.model.add_subsystem('birkhoff', BirkhoffIterGroup(state_options=state_options,
                                                                        time_options=time_options,
                                                                        grid_data=grid_data,
                                                                        ode_class=ode_class))

                    birkhoff = p.model._get_subsystem('birkhoff')

                    birkhoff.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
                    birkhoff.linear_solver = om.DirectSolver()

                    p.setup(force_alloc_complex=True)

                    # Instead of using the TimeComp just transform the node segment taus onto [0, 2]
                    times = grid_data.node_stau + 1

                    solution = np.reshape(times**2 + 2 * times + 1 - 0.5 * np.exp(times), (nn, 1))
                    dsolution_dt = np.reshape(2 * times + 2 - 0.5 * np.exp(times), (nn, 1))

                    p.set_val('birkhoff.initial_states:x', 0.5)
                    p.set_val('birkhoff.ode_all.t', times)
                    p.set_val('birkhoff.ode_all.p', 1.0)

                    # We don't need to provide guesses for these values in this case
                    # p.set_val('birkhoff.final_states:x', solution[-1])
                    # p.set_val('birkhoff.states:x', solution)
                    # p.set_val('birkhoff.state_rates:x', dsolution_dt)

                    p.final_setup()
                    p.run_model()

                    assert_near_equal(solution, p.get_val('birkhoff.states:x'), tolerance=1.0E-9)
                    assert_near_equal(dsolution_dt, p.get_val('birkhoff.state_rates:x'), tolerance=1.0E-9)
                    assert_near_equal(solution[np.newaxis, 0], p.get_val('birkhoff.initial_states:x'), tolerance=1.0E-9)
                    assert_near_equal(solution[np.newaxis, -1], p.get_val('birkhoff.final_states:x'), tolerance=1.0E-9)

                    cpd = p.check_partials(method='cs', compact_print=True, out_stream=None)
                    assert_check_partials(cpd)

    def test_solve_segments_gl_bkwd(self):

        for grid_type in ['lgl', 'cgl']:
            with self.subTest(msg=grid_type):
                with dymos.options.temporary(include_check_partials=True):

                    state_options = {'x': StateOptionsDictionary()}

                    state_options['x']['shape'] = (1,)
                    state_options['x']['units'] = 's**2'
                    state_options['x']['targets'] = ['x']
                    state_options['x']['initial_bounds'] = (None, None)
                    state_options['x']['final_bounds'] = (None, None)
                    state_options['x']['solve_segments'] = 'backward'
                    state_options['x']['rate_source'] = 'x_dot'

                    time_options = TimeOptionsDictionary()
                    grid_data = BirkhoffGrid(num_nodes=31, grid_type=grid_type)
                    nn = grid_data.subset_num_nodes['all']
                    ode_class = SimpleODE

                    p = om.Problem()
                    p.model.add_subsystem('birkhoff', BirkhoffIterGroup(state_options=state_options,
                                                                        time_options=time_options,
                                                                        grid_data=grid_data,
                                                                        ode_class=ode_class))

                    birkhoff = p.model._get_subsystem('birkhoff')

                    birkhoff.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
                    birkhoff.linear_solver = om.DirectSolver()

                    p.setup(force_alloc_complex=True)

                    # Instead of using the TimeComp just transform the node segment taus onto [0, 2]
                    times = grid_data.node_stau + 1

                    solution = np.reshape(times**2 + 2 * times + 1 - 0.5 * np.exp(times), (nn, 1))
                    dsolution_dt = np.reshape(2 * times + 2 - 0.5 * np.exp(times), (nn, 1))

                    p.set_val('birkhoff.final_states:x', solution[-1])
                    p.set_val('birkhoff.ode_all.t', times)
                    p.set_val('birkhoff.ode_all.p', 1.0)

                    # We don't need to provide guesses for these values in this case
                    # p.set_val('birkhoff.final_states:x', solution[-1])
                    # p.set_val('birkhoff.states:x', solution)
                    # p.set_val('birkhoff.state_rates:x', dsolution_dt)

                    p.final_setup()
                    p.run_model()

                    assert_near_equal(solution, p.get_val('birkhoff.states:x'), tolerance=1.0E-9)
                    assert_near_equal(dsolution_dt, p.get_val('birkhoff.state_rates:x'), tolerance=1.0E-9)
                    assert_near_equal(solution[np.newaxis, 0], p.get_val('birkhoff.initial_states:x'), tolerance=1.0E-9)
                    assert_near_equal(solution[np.newaxis, -1], p.get_val('birkhoff.final_states:x'), tolerance=1.0E-9)

                    cpd = p.check_partials(method='cs', compact_print=True, out_stream=None)
                    assert_check_partials(cpd)


if __name__ == '__main__':
    unittest.main()



import unittest

from numpy.testing import assert_almost_equal

from dymos.utils.lg import lg

# Known solutions
x_i = {2: [-0.57735, 0.57735],
       3: [-0.774597, 0.0, 0.774597],
       4: [-0.861136, -0.339981, 0.339981, 0.861136],
       5: [-0.90618, -0.538469, 0.0, 0.538469, 0.90618]}

w_i = {2: [1.0, 1.0],
       3: [0.555556, 0.888889, 0.555556],
       4: [0.347855, 0.652145, 0.652145, 0.347855],
       5: [0.236927, 0.478629, 0.568889, 0.478629, 0.236927]}


class TestLG(unittest.TestCase):

    def test_nodes_and_weights_2(self):
        x_2, w_2 = lg(2)
        assert_almost_equal(x_2, x_i[2], decimal=6)
        assert_almost_equal(w_2, w_i[2], decimal=6)

    def test_nodes_and_weights_3(self):
        x_3, w_3 = lg(3)
        assert_almost_equal(x_3, x_i[3], decimal=6)
        assert_almost_equal(w_3, w_i[3], decimal=6)

    def test_nodes_and_weights_4(self):
        x_4, w_4 = lg(4)
        assert_almost_equal(x_4, x_i[4], decimal=6)
        assert_almost_equal(w_4, w_i[4], decimal=6)

    def test_nodes_and_weights_5(self):
        x_5, w_5 = lg(5)
        assert_almost_equal(x_5, x_i[5], decimal=6)
        assert_almost_equal(w_5, w_i[5], decimal=6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

import openmdao.api as om
from dymos.utils.testing_utils import assert_check_partials

import dymos as dm
from dymos.transcriptions.pseudospectral.components.birkhoff_defect_comp import BirkhoffDefectComp
from dymos.transcriptions.grid_data import BirkhoffGrid

# Modify class so we can run it standalone.
from dymos.utils.misc import CompWrapperConfig


class _PhaseStub():
    """
    A stand-in for the Phase during config_io for testing.
    It just supports the classify_var method and returns "state", the only value needed for unittests.
    """
    def classify_var(self, name):
        return None


CollocationComp = CompWrapperConfig(BirkhoffDefectComp, [_PhaseStub()])


class TestCollocationComp(unittest.TestCase):

    def make_problem(self, grid_type='lgl'):
        dm.options['include_check_partials'] = True

        gd = BirkhoffGrid(num_nodes=21, grid_type=grid_type)
        n = gd.subset_num_nodes['col']
        tau = gd.node_stau
        t = 5 * tau + 5

        self.p = om.Problem(model=om.Group())

        state_options = {'x': {'units': 'm', 'shape': (1,), 'fix_initial': True,
                               'fix_final': False, 'solve_segments': False,
                               'input_initial': False, 'rate_source': 'foo'},
                         'y': {'units': 'm', 'shape': (2, 2), 'fix_initial': False,
                               'fix_final': True, 'solve_segments': False,
                               'input_initial': False, 'rate_source': 'bar'}}

        indep_comp = om.IndepVarComp()
        self.p.model.add_subsystem('indep', indep_comp, promotes_outputs=['*'])

        # Testing the basic ODE xdot = -x, x(0) = 10
        # Solution is x(t) = 10*exp(-t)

        x_val = 10*np.exp(-t)
        y_val = np.zeros((n, 2, 2))
        y_val[:, 0, 0] = x_val
        y_val[:, 1, 1] = x_val

        indep_comp.add_output(
            'dt_dstau',
            val=np.ones(n) * 5, units='s')
        indep_comp.add_output(
            'state_value:x',
            val=x_val, units='m')
        indep_comp.add_output(
            'f_value:x',
            val=-x_val * 5, units='m')
        indep_comp.add_output(
            'f_computed:x',
            val=-x_val, units='m/s')
        indep_comp.add_output(
            'state_value:y',
            val=y_val, units='m')
        indep_comp.add_output(
            'f_value:y',
            val=-y_val * 5, units='m')
        indep_comp.add_output(
            'f_computed:y',
            val=-y_val, units='m/s')

        self.p.model.add_subsystem('defect_comp',
                                   subsys=CollocationComp(grid_data=gd,
                                                          state_options=state_options,
                                                          time_units='s'))

        if grid_type == 'radau-ps':
            src_indices = om.slicer[:-1]
        else:
            src_indices = om.slicer[:]

        self.p.model.connect('state_value:x', 'defect_comp.states:x', src_indices=src_indices)
        self.p.model.connect('f_value:x', 'defect_comp.state_rates:x', src_indices=src_indices)
        self.p.model.connect('state_value:y', 'defect_comp.states:y')
        self.p.model.connect('f_value:y', 'defect_comp.state_rates:y')
        self.p.model.connect('f_computed:x', 'defect_comp.f_computed:x', src_indices=src_indices)
        self.p.model.connect('f_computed:y', 'defect_comp.f_computed:y')
        self.p.model.connect('dt_dstau', 'defect_comp.dt_dstau')

        self.p.setup(force_alloc_complex=True)

        self.p.set_val('defect_comp.initial_states:x', 10.0)
        self.p.set_val('defect_comp.final_states:x', 10 * np.exp(-10))
        self.p.set_val('defect_comp.initial_states:y', np.array([[10.0, 0.0], [0.0, 10.0]]))
        self.p.set_val('defect_comp.final_states:y', np.array([[10 * np.exp(-10), 0.0], [0.0, 10 * np.exp(-10)]]))

        self.p.run_model()

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_results(self):
        for grid_type in ('lgl', 'cgl'):
            with self.subTest(msg=grid_type):
                self.make_problem(grid_type=grid_type)
                assert_almost_equal(self.p['defect_comp.state_defects:x'], 0.0)
                assert_almost_equal(self.p['defect_comp.state_rate_defects:x'], 0.0)
                assert_almost_equal(self.p['defect_comp.state_defects:y'], 0.0)
                assert_almost_equal(self.p['defect_comp.state_rate_defects:y'], 0.0)

    def test_partials(self):
        for grid_type in ('lgl', 'cgl'):
            with self.subTest(msg=grid_type):
                self.make_problem(grid_type=grid_type)
                with np.printoptions(linewidth=1024):
                    cpd = self.p.check_partials(compact_print=True, method='cs', out_stream=None)
                assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import scipy.sparse as sp

import openmdao.api as om

from dymos.transcriptions.grid_data import GridData
from dymos.utils.misc import get_rate_units
from dymos._options import options as dymos_options


class MultipleShootingUpdateComp(om.ExplicitComponent):
    """
    Class definition for the PicardUpdateComp.

    Given the initial state values (for forward propagation) or the final state values
    (for backward propagation), compute the next state value for picard iteration
    using a NonlinearBlockGS solver.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    Attributes
    ----------
    _no_check_partials : bool
        If True, this component will be excluded when partials are checked.
    _M_fwd : array-like
        A selection matrix that pulls appropriate values from x into x_0
        when in forward propagation mode.
    _M_bkwd : array-like
        A selection matrix that pulls appropriate values from x into x_f
        when in backwards propagation mode.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']
        self._M_fwd = None
        self._M_bkwd = None

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare(
            'grid_data', types=GridData,
            desc='Container object for grid info')

        self.options.declare(
            'state_options', types=dict,
            desc='Dictionary of state names/options for the phase')

        self.options.declare(
            'time_units', default=None, allow_none=True, types=str,
            desc='Units of time')

    def configure_io(self, phase):
        """
        I/O creation is delayed until configure so we can determine shape and units.

        Parameters
        ----------
        phase : Phase
            The phase object that contains this collocation comp.
        """
        gd = self.options['grid_data']
        num_nodes = gd.subset_num_nodes['all']
        num_segs = gd.num_segments
        time_units = self.options['time_units']
        state_options = self.options['state_options']

        # Construct the forward and backward mapping matrices
        seg_start_nodes = gd.subset_node_indices['segment_ends'][::2]
        seg_end_nodes = gd.subset_node_indices['segment_ends'][1::2]

        M_fwd = sp.lil_array((num_segs, num_nodes), dtype=int)
        for iseg in range(1, gd.num_segments):
            # The ith row of M_fwd contains in the column pertaining to first node in the (i-1)th segment
            M_fwd[iseg, seg_end_nodes[iseg-1]] = 1

        M_bkwd = sp.lil_array((num_segs, num_nodes), dtype=int)
        for iseg in range(gd.num_segments-1):
            # The ith row of M_bkwd contains in the column pertaining to first node in the (i+1)th segment
            M_bkwd[iseg, seg_start_nodes[iseg+1]] = 1

        self._M_fwd = {}
        self._M_bkwd = {}

        self._var_names = var_names = {}
        for state_name, options in state_options.items():
            var_names[state_name] = {
                'x_a': f'initial_states:{state_name}',
                'x_b': f'final_states:{state_name}',
                'x_0': f'seg_initial_states:{state_name}',
                'x_f': f'seg_final_states:{state_name}',
                'x': f'states:{state_name}'
            }

        for state_name, options in state_options.items():
            shape = options['shape']
            units = options['units']
            size = np.prod(shape)

            rate_units = get_rate_units(units, time_units)
            var_names = self._var_names[state_name]

            self.add_input(
                name=var_names['x'],
                shape=(num_nodes,) + shape,
                desc='State value at the nodes.',
                units=rate_units)

            if options['solve_segments'] == 'forward':
                self._M_fwd[state_name] = sp.kron(M_fwd, sp.eye(size, dtype=int))

                self.add_input(
                    name=var_names['x_a'],
                    shape=(1,) + shape,
                    desc=f'Desired initial value of state {state_name} in phase.',
                    units=units
                )
                self.add_output(
                    name=var_names['x_0'],
                    shape=(num_segs,) + shape,
                    val=0.0,
                    desc=f'Given initial value of state {state_name} in each segment.',
                    units=units
                )

                ar_size_x_nn0 = np.arange(size, dtype=int)
                self.declare_partials(of=var_names['x_0'],
                                      wrt=var_names['x_a'],
                                      rows=ar_size_x_nn0,
                                      cols=np.zeros_like(ar_size_x_nn0),
                                      val=1.0)

                if num_segs > 1:
                    rs, cs = self._M_fwd[state_name].nonzero()
                    self.declare_partials(of=var_names['x_0'],
                                          wrt=var_names['x'],
                                          rows=rs, cols=cs,
                                          val=self._M_fwd[state_name].data.ravel())

            elif options['solve_segments'] == 'backward':
                self._M_bkwd[state_name] = sp.kron(M_bkwd, sp.eye(size, dtype=int))

                self.add_input(
                    name=var_names['x_b'],
                    shape=(1,) + shape,
                    desc=f'Desired initial value of state {state_name} in phase.',
                    units=units
                )

                self.add_output(
                    name=var_names['x_f'],
                    shape=(num_segs,) + shape,
                    val=0.0,
                    desc=f'Given final value of state {state_name} in each segment.',
                    units=units
                )

                ar_size_x_nnf = (num_segs - 1) * size + np.arange(size, dtype=int)
                self.declare_partials(of=var_names['x_f'],
                                      wrt=var_names['x_b'],
                                      rows=ar_size_x_nnf,
                                      cols=np.zeros_like(ar_size_x_nnf),
                                      val=1.0)

                if num_segs > 1:
                    rs, cs = self._M_bkwd[state_name].nonzero()
                    self.declare_partials(of=var_names['x_f'],
                                          wrt=var_names['x'],
                                          rows=rs, cols=cs,
                                          val=self._M_bkwd[state_name].data.ravel())

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute component outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : dict or None
            If not None, dict containing discrete input values.
        discrete_outputs : dict or None
            If not None, dict containing discrete output values.
        """
        gd = self.options['grid_data']
        nn = gd.num_nodes

        for state_name, options in self.options['state_options'].items():
            var_names = self._var_names[state_name]
            x = inputs[var_names['x']]
            x_flat = x.reshape(nn, -1)

            if options['solve_segments'] == 'forward':
                M_fwd = self._M_fwd[state_name]
                x_a = inputs[var_names['x_a']]
                outputs[var_names['x_0']] = M_fwd @ x_flat
                outputs[var_names['x_0']][0, ...] = x_a[0, ...]
                # print(x_flat)
                # print(x_a)
                # print(outputs[var_names['x_0']])

            elif options['solve_segments'] == 'backward':
                M_bkwd = self._M_bkwd[state_name]
                x_b = inputs[var_names['x_b']]
                outputs[var_names['x_f']] = M_bkwd @ x_flat
                outputs[var_names['x_f']][-1, ...] = x_b

            else:
                raise ValueError(f'{self.msginfo}: Invalid direction of integration: {options["solve_segments"]}')



import unittest
import warnings

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


try:
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
except ImportError:
    plt = None


OPTIMIZER = 'SLSQP'


class _A(object):
    pass


class _B(om.Group):
    pass


class _C(om.ExplicitComponent):
    pass


class _D(om.ExplicitComponent):
    ode_options = None


@use_tempdirs
class TestPhaseBase(unittest.TestCase):

    def test_invalid_ode_wrong_class(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=_A,
                         transcription=dm.GaussLobatto(num_segments=20, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=False)

        phase.add_parameter('g', units='m/s**2', opt=True, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.linear_solver = om.DirectSolver()

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = 'If given as a class, ode_class must be derived from openmdao.core.System.'
        self.assertEqual(expected, str(e.exception))

    def test_invalid_ode_instance(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=_A(),
                         transcription=dm.GaussLobatto(num_segments=20, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)
        phase.add_state('y', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=False)

        phase.add_parameter('g', units='m/s**2', opt=True, val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.linear_solver = om.DirectSolver()

        with self.assertRaises(ValueError) as e:
            p.setup(check=True)

        expected = 'ode_class must be given as a callable object that returns an object derived ' \
                   'from openmdao.core.System, or as a class derived from openmdao.core.System.'
        self.assertEqual(expected, str(e.exception))

    def test_add_existing_parameter_as_parameter(self):

        p = dm.Phase(ode_class=_A,
                     transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.add_parameter('theta')

        with self.assertRaises(ValueError) as e:
            p.add_parameter('theta')

        expected = 'theta has already been added as a parameter.'
        self.assertEqual(str(e.exception), expected)

    def test_add_existing_control_as_parameter(self):

        p = dm.Phase(ode_class=BrachistochroneODE,
                     transcription=dm.GaussLobatto(num_segments=8, order=3))

        p.add_control('theta')

        with self.assertRaises(ValueError) as e:
            p.add_parameter('theta')

        expected = 'theta has already been added as a control.'
        self.assertEqual(str(e.exception), expected)

    def test_invalid_options_nonoptimal_param(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()
        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=16, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)

        phase.add_state('y', fix_initial=True, fix_final=False)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, lower=5, upper=10,
                            ref0=5, ref=10, scaler=1, adder=0)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.linear_solver = om.DirectSolver()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            p.setup(check=False)

        print('\n'.join([str(ww.message) for ww in w]))

        expected = 'Invalid options for non-optimal parameter \'g\' in phase \'phase0\': ' \
                   'lower, upper, scaler, adder, ref, ref0'

        self.assertIn(expected, [str(ww.message) for ww in w])

    def test_invalid_options_nonoptimal_control(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=False,
                          units='deg', lower=0.01, upper=179.9, scaler=1, ref=1, ref0=0)

        phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.linear_solver = om.DirectSolver()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            p.setup(check=True)

        expected = 'Invalid options for non-optimal control \'theta\' in phase \'phase0\': ' \
                   'lower, upper, scaler, ref, ref0'

        self.assertIn(expected, [str(ww.message) for ww in w])

    def test_invalid_boundary_loc(self):

        p = dm.Phase(ode_class=BrachistochroneODE,
                     transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        with self.assertRaises(ValueError) as e:
            p.add_boundary_constraint('x', loc='foo')

        expected = 'Invalid boundary constraint location "foo". Must be "initial" or "final".'
        self.assertEqual(str(e.exception), expected)

    def test_objective_parameter_gl(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=8, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(p['phase0.t_duration'], 10, tolerance=1.0E-3)

    def test_objective_parameter_radau(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.Radau(num_segments=20, order=3, compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(4, 10))

        phase.add_state('x', fix_initial=True, fix_final=True)

        phase.add_state('y', fix_initial=True, fix_final=True)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, opt=True,
                          rate_continuity_scaler=1.0,
                          units='deg', lower=0.01, upper=179.9, ref=1, ref0=0)

        phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

        # Minimize time at the end of the phase
        phase.add_objective('g')

        p.model.options['assembled_jac_type'] = 'csc'
        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(p['phase0.t_duration'], 10, tolerance=1.0E-3)

    def test_control_boundary_constraint_gl(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=20,
                                                       order=3,
                                                       compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='s')

        phase.add_state('x', fix_initial=True, fix_final=False)

        phase.add_state('y', fix_initial=True, fix_final=False)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('theta', loc='final', lower=90.0, upper=90.0, units='deg')

        # Minimize time at the end of the phase
        phase.add_objective('time')

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(p.get_val('phase0.timeseries.theta', units='deg')[-1], 90.0)

    def test_control_rate_boundary_constraint_gl(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=20,
                                                       order=3,
                                                       compressed=True))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(0.1, 10))

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)

        phase.add_state('y', fix_initial=True, fix_final=False)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, units='deg',
                          lower=0.01, upper=179.9)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('theta_rate', loc='final', equals=0.0, units='deg/s')

        # Minimize time at the end of the phase
        phase.add_objective('time')

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 8.0)

        failed = p.run_driver()

        self.assertFalse(failed)
        assert_near_equal(p.get_val('phase0.timeseries.theta_rate')[-1, ...], 0.0, tolerance=1.0E-3)

    def test_control_rate2_boundary_constraint_gl(self):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneODE,
                         transcription=dm.GaussLobatto(num_segments=20,
                                                       order=3,
                                                       compressed=False))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, continuity_scaler=1.0)

        phase.add_state('y', fix_initial=True, fix_final=False, continuity_ref=1.0)

        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True, rate2_continuity=True,
                          rate_continuity_scaler=0.01, rate2_continuity_scaler=0.01,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', opt=False, units='m/s**2', val=9.80665)

        phase.add_boundary_constraint('theta_rate2', loc='final', equals=0.0, units='deg/s**2')

        # Minimize time at the end of the phase
        phase.add_objective('time')

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 8.0)

        failed = p.run_driver()

        self.assertFalse(failed)
        assert_near_equal(p.get_val('phase0.timeseries.theta_rate2')[-1, ...], 0.0, tolerance=1.0E-3)

    def test_parameter_multiple_boundary_constraints(self):

        expected = 'In phase phase0, parameter `g` is subject to multiple boundary ' \
                   'or path constraints.\nParameters are single values that do not change in ' \
                   'time, and may only be used in a single boundary or path constraint.'

        transcriptions = {'gauss-lobatto': dm.GaussLobatto,
                          'radau-ps': dm.Radau,
                          'explicit-shooting': dm.ExplicitShooting}

        for txname, tx in transcriptions.items():
            with self.subTest(msg=f'{txname}'):
                p = om.Problem(model=om.Group())

                p.driver = om.ScipyOptimizeDriver()
                p.driver.declare_coloring()
                if tx is dm.ExplicitShooting:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(grid=dm.GaussLobattoGrid(num_segments=5, nodes_per_seg=3)))
                else:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(num_segments=5, order=3))

                p.model.add_subsystem('phase0', phase)

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                phase.add_state('x', fix_initial=True, fix_final=False, continuity_scaler=1.0)

                phase.add_state('y', fix_initial=True, fix_final=False, continuity_ref=1.0)

                phase.add_state('v', fix_initial=True, fix_final=False)

                phase.add_control('theta', continuity=True, rate_continuity=True, rate2_continuity=True,
                                  rate_continuity_ref=100., rate2_continuity_scaler=0.01,
                                  units='deg', lower=0.01, upper=179.9)

                phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

                # We'll let g vary, but make sure it hits the desired value.
                # It's a static parameter, so it shouldn't matter whether we enforce it
                # at the start or the end of the phase, so here we'll do both.
                # Note if we make these equality constraints, some optimizers (SLSQP) will
                # see the problem as infeasible.
                phase.add_boundary_constraint('x', loc='final', units='m', equals=10)
                phase.add_boundary_constraint('y', loc='final', units='m', equals=5)
                phase.add_boundary_constraint('g', loc='initial', units='m/s**2', upper=9.80665)
                phase.add_boundary_constraint('g', loc='final', units='m/s**2', upper=9.80665)

                # Minimize time at the end of the phase
                phase.add_objective('time_phase', loc='final', scaler=10)

                p.model.linear_solver = om.DirectSolver()

                with self.assertRaises(RuntimeError) as e:
                    p.setup()

                self.assertEqual(str(e.exception), expected)

    def test_parameter_initial_boundary_constraint(self):

        transcriptions = {'gauss-lobatto': dm.GaussLobatto,
                          'radau-ps': dm.Radau,
                          'explicit-shooting': dm.ExplicitShooting}

        for txname, tx in transcriptions.items():
            with self.subTest(msg=f'{txname}'):
                p = om.Problem(model=om.Group())

                p.driver = om.ScipyOptimizeDriver()
                p.driver.declare_coloring()

                if tx is dm.ExplicitShooting:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(grid=dm.GaussLobattoGrid(num_segments=5, nodes_per_seg=3)))
                else:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(num_segments=5, order=3))

                p.model.add_subsystem('phase0', phase)

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                phase.add_state('x', fix_initial=True, fix_final=False, continuity_scaler=1.0)

                phase.add_state('y', fix_initial=True, fix_final=False, continuity_ref=1.0)

                phase.add_state('v', fix_initial=True, fix_final=False)

                phase.add_control('theta', continuity=True, rate_continuity=True, rate2_continuity=False,
                                  rate_continuity_scaler=0.01, rate2_continuity_ref=1.0,
                                  units='deg', lower=0.01, upper=179.9)

                phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

                # We'll let g vary, but make sure it hits the desired value.
                # It's a static parameter, so it shouldn't matter whether we enforce it
                # at the start or the end of the phase.
                # Note if we make these equality constraints, some optimizers (SLSQP) will
                # see the problem as infeasible.
                phase.add_boundary_constraint('x', loc='final', units='m', equals=10)
                phase.add_boundary_constraint('y', loc='final', units='m', equals=5)
                phase.add_boundary_constraint('g', loc='initial', units='m/s**2', upper=9.80665)

                # Minimize time at the end of the phase
                phase.add_objective('time_phase', loc='final', scaler=10)

                p.model.linear_solver = om.DirectSolver()
                p.setup(check=True)

                phase.set_time_val(initial=0, duration=2.0)
                phase.set_state_val('x', (0, 10))
                phase.set_state_val('y', (10, 5))
                phase.set_state_val('v', (0, 9.9))
                phase.set_control_val('theta', (5, 100))
                phase.set_parameter_val('g', 5.0)

                p.run_driver()

                assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016,
                                  tolerance=1.0E-4)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[0], 9.80665,
                                  tolerance=1.0E-6)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[-1], 9.80665,
                                  tolerance=1.0E-6)

    def test_parameter_final_boundary_constraint(self):

        transcriptions = {'gauss-lobatto': dm.GaussLobatto,
                          'radau-ps': dm.Radau,
                          'explicit-shooting': dm.ExplicitShooting}

        for txname, tx in transcriptions.items():
            with self.subTest(msg=f'{txname}'):
                p = om.Problem(model=om.Group())

                p.driver = om.ScipyOptimizeDriver()
                p.driver.declare_coloring()

                if isinstance(tx, dm.ExplicitShooting):
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(grid=dm.GaussLobattoGrid(num_segments=5, nodes_per_seg=3)))
                else:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(num_segments=5, order=3))

                p.model.add_subsystem('phase0', phase)

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                phase.add_state('x', fix_initial=True, fix_final=False)

                phase.add_state('y', fix_initial=True, fix_final=False)

                phase.add_state('v', fix_initial=True, fix_final=False)

                phase.add_control('theta', continuity=True, rate_continuity=True,
                                  units='deg', lower=0.01, upper=179.9)

                phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

                # We'll let g vary, but make sure it hits the desired value.
                # It's a static parameter, so it shouldn't matter whether we enforce it
                # at the start or the end of the phase.
                # Note if we make these equality constraints, some optimizers (SLSQP) will
                # see the problem as infeasible.
                phase.add_boundary_constraint('x', loc='final', units='m', equals=10)
                phase.add_boundary_constraint('y', loc='final', units='m', equals=5)
                phase.add_boundary_constraint('g', loc='final', units='m/s**2', upper=9.80665)

                # Minimize time at the end of the phase
                phase.add_objective('time_phase', loc='final', scaler=10)

                p.model.linear_solver = om.DirectSolver()
                p.setup(check=True)

                phase.set_time_val(initial=0, duration=2.0)
                phase.set_state_val('x', (0, 10))
                phase.set_state_val('y', (10, 5))
                phase.set_state_val('v', (0, 9.9))
                phase.set_control_val('theta', (5, 100))
                phase.set_parameter_val('g', 5.0)

                p.run_driver()

                assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016,
                                  tolerance=1.0E-4)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[0], 9.80665,
                                  tolerance=1.0E-6)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[-1], 9.80665,
                                  tolerance=1.0E-6)

    def test_parameter_path_constraint(self):

        transcriptions = {'gauss-lobatto': dm.GaussLobatto,
                          'radau-ps': dm.Radau,
                          'explicit-shooting': dm.ExplicitShooting}

        for txname, tx in transcriptions.items():
            with self.subTest(msg=f'{txname}'):
                p = om.Problem(model=om.Group())

                p.driver = om.ScipyOptimizeDriver()
                p.driver.declare_coloring()

                if isinstance(tx, dm.ExplicitShooting):
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(grid=dm.GaussLobattoGrid(num_segments=5, order=3)))
                else:
                    phase = dm.Phase(ode_class=BrachistochroneODE,
                                     transcription=tx(num_segments=5, order=3))

                p.model.add_subsystem('phase0', phase)

                phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

                phase.add_state('x', fix_initial=True, fix_final=False)

                phase.add_state('y', fix_initial=True, fix_final=False)

                phase.add_state('v', fix_initial=True, fix_final=False)

                phase.add_control('theta', continuity=True, rate_continuity=True, rate2_continuity=False,
                                  units='deg', lower=0.01, upper=179.9)

                phase.add_parameter('g', opt=True, units='m/s**2', val=9.80665)

                # We'll let g vary, but make sure it hits the desired value.
                # It's a static parameter, so it shouldn't matter whether we enforce it
                # at the start or the end of the phase.  In this case we'll use a path
                # constraint to enforce it, but it will still just be a single scalar constraint.
                phase.add_boundary_constraint('x', loc='final', units='m', equals=10)
                phase.add_boundary_constraint('y', loc='final', units='m', equals=5)
                phase.add_path_constraint('g', units='m/s**2', upper=9.80665)

                # Minimize time at the end of the phase
                phase.add_objective('time_phase', loc='final', scaler=10)

                p.model.linear_solver = om.DirectSolver()
                p.setup(check=True)

                phase.set_time_val(initial=0, duration=2.0)
                phase.set_state_val('x', (0, 10))
                phase.set_state_val('y', (10, 5))
                phase.set_state_val('v', (0, 9.9))
                phase.set_control_val('theta', (5, 100))
                phase.set_parameter_val('g', 5.0)

                p.run_driver()

                assert_near_equal(p.get_val('phase0.timeseries.time')[-1], 1.8016,
                                  tolerance=1.0E-4)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[0], 9.80665,
                                  tolerance=1.0E-6)
                assert_near_equal(p.get_val('phase0.parameter_vals:g')[-1], 9.80665,
                                  tolerance=1.0E-6)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest
import pathlib

import numpy as np

try:
    import matplotlib
except ImportError:
    matplotlib = None

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.examples.battery_multibranch.battery_multibranch_ode import BatteryODE
from dymos.utils.lgl import lgl
from dymos.utils.misc import om_version
from dymos.utils.testing_utils import _get_reports_dir
from dymos.visualization.timeseries_plots import timeseries_plots


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestTimeSeriesPlotsBasics(unittest.TestCase):

    def setUp(self):
        optimizer = 'SLSQP'
        num_segments = 8
        transcription_order = 3
        compressed = True

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()

        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)
        phase.timeseries_options['include_control_rates'] = True
        traj.add_phase('phase0', phase)

        p.model.add_subsystem('traj0', traj)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False)
        phase.add_state('y', fix_initial=True, fix_final=False)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', val=9.80665)

        phase.add_timeseries('timeseries2',
                             transcription=dm.Radau(num_segments=num_segments * 5,
                                                    order=transcription_order,
                                                    compressed=compressed),
                             subset='control_input')

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=['unconnected_inputs'])

        phase.set_time_val(initial=0.0, duration=2.0)
        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.setup()

        self.p = p

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_timeseries_plots(self):
        with dm.options.temporary(plots='matplotlib'):
            dm.run_problem(self.p, make_plots=False)

            sol_db = 'dymos_solution.db'
            if om_version()[0] > (3, 34, 2):
                sol_db = self.p.get_outputs_dir() / sol_db

            timeseries_plots(sol_db, problem=self.p)
            plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

            self.assertTrue(plot_dir.joinpath('x.png').exists())
            self.assertTrue(plot_dir.joinpath('y.png').exists())
            self.assertTrue(plot_dir.joinpath('v.png').exists())
            self.assertTrue(plot_dir.joinpath('theta.png').exists())
            self.assertTrue(plot_dir.joinpath('theta_rate.png').exists())
            self.assertTrue(plot_dir.joinpath('theta_rate2.png').exists())

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_timeseries_plots_solution_only_set_solution_record_file(self):
        temp = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        # records to the default file 'dymos_simulation.db'
        dm.run_problem(self.p, make_plots=False, solution_record_file='solution_record_file.db')

        sol_db = 'solution_record_file.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = self.p.get_outputs_dir() / sol_db

        timeseries_plots(sol_db, problem=self.p)
        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath('plots')

        self.assertTrue(plot_dir.joinpath('x.png').exists())
        self.assertTrue(plot_dir.joinpath('y.png').exists())
        self.assertTrue(plot_dir.joinpath('v.png').exists())
        self.assertTrue(plot_dir.joinpath('theta.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate2.png').exists())

        dm.options['plots'] = temp

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_timeseries_plots_solution_and_simulation(self):
        temp = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        dm.run_problem(self.p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        sol_db = 'dymos_solution.db'
        sim_db = 'simulation_record_file.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = self.p.get_outputs_dir() / sol_db
            sim_db = self.p.model.traj0.sim_prob.get_outputs_dir() / sim_db

        timeseries_plots(sol_db,
                         simulation_record_file=sim_db,
                         problem=self.p)

        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath("plots").resolve()
        self.assertTrue(plot_dir.joinpath('x.png').exists())
        self.assertTrue(plot_dir.joinpath('y.png').exists())
        self.assertTrue(plot_dir.joinpath('v.png').exists())
        self.assertTrue(plot_dir.joinpath('theta.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate2.png').exists())

        dm.options['plots'] = temp

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_brachistochrone_timeseries_plots_set_plot_dir(self):

        temp = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        dm.run_problem(self.p, make_plots=False)

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = self.p.get_outputs_dir() / sol_db

        plot_dir = pathlib.Path(_get_reports_dir(self.p)).joinpath("test_plot_dir").resolve()
        timeseries_plots(sol_db, plot_dir=plot_dir)

        self.assertTrue(plot_dir.joinpath('x.png').exists())
        self.assertTrue(plot_dir.joinpath('y.png').exists())
        self.assertTrue(plot_dir.joinpath('v.png').exists())
        self.assertTrue(plot_dir.joinpath('theta.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate.png').exists())
        self.assertTrue(plot_dir.joinpath('theta_rate2.png').exists())

        dm.options['plots'] = temp


@use_tempdirs
class TestTimeSeriesPlotsMultiPhase(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_trajectory_linked_phases_make_plot(self):
        temp = dm.options['plots']

        dm.options['plots'] = 'matplotlib'

        self.traj = dm.Trajectory()
        p = self.p = om.Problem(model=self.traj)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.declare_coloring()

        # First Phase (burn)
        burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=4,
                                                                                order=3))

        self.traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn1.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Second Phase (Coast)

        coast = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=10,
                                                                                order=3))

        self.traj.add_phase('coast', coast)

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), duration_ref=10,
                               units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', targets=['r'], units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True, ref=1.0E-12, defect_ref=1.0E-12,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        coast.add_control('u1', targets=['u1'], opt=False, val=0.0, units='deg')
        coast.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        # Third Phase (burn)

        burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=3,
                                                                                order=3))

        self.traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), initial_ref=10,
                               units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_ref=1.0E-6,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn2.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn2.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU')

        burn2.add_objective('deltav', loc='final')

        # Link Phases
        self.traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                              vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])
        self.traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

        # Finish Problem Setup
        p.model.linear_solver = om.DirectSolver()

        for phase_name, phase in self.traj._phases.items():
            phase.timeseries_options['use_prefix'] = True
            phase.timeseries_options['include_state_rates'] = True
            phase.timeseries_options['include_t_phase'] = True
            phase.timeseries_options['include_control_rates'] = True

        p.setup(check=True)

        # Set Initial Guesses
        burn1.set_time_val(initial=0.0, duration=2.25)
        burn1.set_state_val('r', [1, 1.5])
        burn1.set_state_val('theta', [0, 1.7])
        burn1.set_state_val('vr', [0, 0])
        burn1.set_state_val('vt', [1, 1])
        burn1.set_state_val('accel', [0.1, 0])
        burn1.set_state_val('deltav', [0, 0.1])
        burn1.set_control_val('u1', [-3.5, 13.0])
        burn1.set_parameter_val('c', 1.5)

        coast.set_time_val(initial=2.25, duration=3.0)
        coast.set_state_val('r', [1.3, 1.5])
        coast.set_state_val('theta', [2.1767, 1.7])
        coast.set_state_val('vr', [0.3285, 0.0])
        coast.set_state_val('vt', [0.97, 1])
        coast.set_state_val('accel', [0.0, 0.0])
        coast.set_parameter_val('c', 1.5)

        burn2.set_time_val(initial=5.25, duration=1.75)
        burn2.set_state_val('r', [1, 3])
        burn2.set_state_val('theta', [0, 4])
        burn2.set_state_val('vr', [0, 0])
        burn2.set_state_val('vt', [1, np.sqrt(1./3.)])
        burn2.set_state_val('accel', [0.1, 0])
        burn2.set_state_val('deltav', [0.1, 0.2])
        burn2.set_control_val('u1', [1.0, 1.0])
        burn2.set_parameter_val('c', 1.5)

        dm.run_problem(p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        sol_db = 'dymos_solution.db'
        sim_db = 'simulation_record_file.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = self.p.get_outputs_dir() / sol_db
            sim_db = self.p.model.sim_prob.get_outputs_dir() / sim_db

        timeseries_plots(sol_db,
                         simulation_record_file=sim_db,
                         problem=p)
        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath("plots")

        for varname in ['time_phase', 'states:r', 'state_rates:r', 'states:theta',
                        'state_rates:theta', 'states:vr', 'state_rates:vr', 'states:vt',
                        'state_rates:vt', 'states:accel',
                        'state_rates:accel', 'states:deltav', 'state_rates:deltav',
                        'controls:u1', 'control_rates:u1_rate', 'control_rates:u1_rate2']:
            plotfile = plot_dir.joinpath(varname.replace(":", "_") + '.png')
            self.assertTrue(plotfile.exists(), msg=f'{plotfile} does not exist!')

        dm.options['plots'] = temp

    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_overlapping_phases_make_plot(self):

        _temp = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        prob = om.Problem()

        opt = prob.driver = om.ScipyOptimizeDriver()
        opt.declare_coloring()
        opt.options['optimizer'] = 'SLSQP'

        num_seg = 5
        seg_ends, _ = lgl(num_seg + 1)

        traj = prob.model.add_subsystem('traj', dm.Trajectory())

        # First phase: normal operation.
        transcription = dm.Radau(num_segments=num_seg, order=5, segment_ends=seg_ends,
                                 compressed=False)
        phase0 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p0 = traj.add_phase('phase0', phase0)

        traj_p0.set_time_options(fix_initial=True, fix_duration=True)
        traj_p0.add_state('state_of_charge', fix_initial=True, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase: normal operation.

        phase1 = dm.Phase(ode_class=BatteryODE, transcription=transcription)
        traj_p1 = traj.add_phase('phase1', phase1)

        traj_p1.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1.add_state('state_of_charge', fix_initial=False, fix_final=False,
                          targets=['SOC'], rate_source='dXdt:SOC')
        traj_p1.add_objective('time', loc='final')

        # Second phase, but with battery failure.

        phase1_bfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_battery': 2},
                                transcription=transcription)
        traj_p1_bfail = traj.add_phase('phase1_bfail', phase1_bfail)

        traj_p1_bfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_bfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        # Second phase, but with motor failure.

        phase1_mfail = dm.Phase(ode_class=BatteryODE, ode_init_kwargs={'num_motor': 2},
                                transcription=transcription)
        traj_p1_mfail = traj.add_phase('phase1_mfail', phase1_mfail)

        traj_p1_mfail.set_time_options(fix_initial=False, fix_duration=True)
        traj_p1_mfail.add_state('state_of_charge', fix_initial=False, fix_final=False,
                                targets=['SOC'], rate_source='dXdt:SOC')

        traj.link_phases(phases=['phase0', 'phase1'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_bfail'], vars=['state_of_charge', 'time'])
        traj.link_phases(phases=['phase0', 'phase1_mfail'], vars=['state_of_charge', 'time'])

        prob.model.options['assembled_jac_type'] = 'csc'
        prob.model.linear_solver = om.DirectSolver(assemble_jac=True)

        for phase_name, phase in traj._phases.items():
            phase.timeseries_options['include_t_phase'] = True
            phase.timeseries_options['include_state_rates'] = True
            phase.timeseries_options['include_control_rates'] = True

        prob.setup()

        phase0.set_time_val(initial=0.0, duration=3600)
        phase1.set_time_val(initial=3600.0, duration=3600.0)
        phase1_bfail.set_time_val(initial=3600.0, duration=3600.0)
        phase1_mfail.set_time_val(initial=3600.0, duration=3600.0)
        phase0.set_state_val('state_of_charge', 1.0)

        prob.set_solver_print(level=0)
        dm.run_problem(prob, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        sol_db = 'dymos_solution.db'
        sim_db = 'simulation_record_file.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = prob.get_outputs_dir() / sol_db
            sim_db = prob.model.traj.sim_prob.get_outputs_dir() / sim_db

        plot_dir = pathlib.Path(_get_reports_dir(prob)).joinpath("plots")

        timeseries_plots(sol_db,
                         simulation_record_file=sim_db,
                         problem=prob)

        self.assertTrue(plot_dir.joinpath('time_phase.png').exists())
        self.assertTrue(plot_dir.joinpath('state_of_charge.png').exists())
        self.assertTrue(plot_dir.joinpath('state_of_charge.png').exists())

        dm.options['plots'] = _temp

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipIf(matplotlib is None, "This test requires matplotlib")
    def test_trajectory_linked_phases_make_plot_missing_data(self):
        """
        Test that plots are still generated even if the phases don't share the exact same
        variables in the timeseries.
        """
        temp = dm.options['plots']
        dm.options['plots'] = 'matplotlib'

        self.traj = dm.Trajectory()
        p = self.p = om.Problem(model=self.traj)

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.declare_coloring()

        # First Phase (burn)
        burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=4,
                                                                                order=3))

        self.traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn1.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU', include_timeseries=True)

        # Second Phase (Coast)

        coast = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=10,
                                                                                order=3))

        self.traj.add_phase('coast', coast)

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), duration_ref=10,
                               units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', targets=['r'], units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True, ref=1.0E-12, defect_ref=1.0E-12,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        coast.add_parameter('u1', targets=['u1'], opt=False, val=0.0, units='deg', include_timeseries=True)
        coast.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU', include_timeseries=True)

        # Third Phase (burn)

        burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=3,
                                                                                order=3))

        self.traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), initial_ref=10,
                               units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_ref=1.0E-6,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn2.add_control('u1', targets=['u1'], rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn2.add_parameter('c', opt=False, val=1.5, targets=['c'], units='DU/TU', include_timeseries=True)

        burn2.add_objective('deltav', loc='final')

        # Link Phases
        self.traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                              vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])
        self.traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

        # Finish Problem Setup
        p.model.linear_solver = om.DirectSolver()

        for phase_name, phase in self.traj._phases.items():
            phase.timeseries_options['include_t_phase'] = True
            phase.timeseries_options['include_state_rates'] = True
            phase.timeseries_options['include_control_rates'] = True

        p.setup(check=True)

        # Set Initial Guesses
        burn1.set_time_val(initial=0.0, duration=2.25)
        burn1.set_state_val('r', [1, 1.5])
        burn1.set_state_val('theta', [0, 1.7])
        burn1.set_state_val('vr', [0, 0])
        burn1.set_state_val('vt', [1, 1])
        burn1.set_state_val('accel', [0.1, 0])
        burn1.set_state_val('deltav', [0, 0.1])
        burn1.set_control_val('u1', [-3.5, 13.0])
        burn1.set_parameter_val('c', 1.5)

        coast.set_time_val(initial=2.25, duration=3.0)
        coast.set_state_val('r', [1.3, 1.5])
        coast.set_state_val('theta', [2.1767, 1.7])
        coast.set_state_val('vr', [0.3285, 0.0])
        coast.set_state_val('vt', [0.97, 1])
        coast.set_state_val('accel', [0.0, 0.0])
        coast.set_parameter_val('c', 1.5)

        burn2.set_time_val(initial=5.25, duration=1.75)
        burn2.set_state_val('r', [1, 3])
        burn2.set_state_val('theta', [0, 4])
        burn2.set_state_val('vr', [0, 0])
        burn2.set_state_val('vt', [1, np.sqrt(1./3.)])
        burn2.set_state_val('accel', [0.1, 0])
        burn2.set_state_val('deltav', [0.1, 0.2])
        burn2.set_control_val('u1', [1.0, 1.0])
        burn2.set_parameter_val('c', 1.5)

        dm.run_problem(p, simulate=True, make_plots=False,
                       simulation_record_file='simulation_record_file.db')

        sol_db = 'dymos_solution.db'
        sim_db = 'simulation_record_file.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.sim_prob.get_outputs_dir() / sim_db

        timeseries_plots(sol_db, sim_db, problem=p)
        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath("plots")

        for varname in ['time_phase', 'r', 'r_dot', 'theta',
                        'theta_dot', 'vr', 'vr_dot', 'vt',
                        'vt_dot', 'accel',
                        'at_dot', 'deltav', 'deltav_dot',
                        'u1', 'u1_rate', 'u1_rate2']:
            self.assertTrue(plot_dir.joinpath(varname + '.png').exists(),
                            msg=varname + '.png does not exist!')

        dm.options['plots'] = temp


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import scipy.special as sp


def birkhoff_matrix(tau, w, grid_type):
    """
    Returns the pseudospectral integration matrix for a Birkhoff polynomial at the given nodes.

    Parameters
    ----------
    tau : ndarray[:]
        Vector of given nodes in the polynomial.
    w : ndarray[:]
        Vector of nodes at which the polynomial is evaluated.
    grid_type : str
        The type of Gaussian grid used in the transcription.

    Returns
    -------
    np.array
        An num_disc_nodes-1 x num_disc_nodes matrix used for the interpolation of state values
        at the interior LGL nodes.
    """
    N = tau.size - 1
    end_node = N if tau[-1] == 1 else N + 1

    alpha = np.zeros((N + 1, N + 1))
    S = np.zeros((N + 1, N + 1))

    if grid_type[0] == 'l':
        pol = sp.eval_legendre
    elif grid_type[0] == 'c':
        pol = sp.eval_chebyt
    else:
        raise ValueError('invalid grid type')

    for j in range(0, N + 1):
        for n in range(0, end_node):
            alpha[n, j] = w[j] * pol(n, tau[j])
        if grid_type == 'lgl':
            alpha[N, j] = N * pol(N, tau[j]) * w[j] / (2 * N + 1)
        elif grid_type == 'cgl':
            alpha[N, j] = pol(N, tau[j]) * w[j] / 2

    if grid_type[0] == 'l':
        for i in range(0, N + 1):
            S[i, 0] = (tau[i] - tau[0]) / 2

            for n in range(1, N+1):
                gamma = 2 / (2 * n + 1)
                int_p = (pol(n+1, tau[i]) - pol(n-1, tau[i])) / (2*n+1)
                S[i, n] = int_p / gamma

    elif grid_type[0] == 'c':
        gamma = np.pi / 2
        for i in range(0, N+1):
            # chebyshev polynomial of order 0: 1
            S[i, 0] = (tau[i] - tau[0]) / np.pi

            # chebyshev polynomial of order 1: x
            S[i, 1] = (tau[i]**2 - tau[0]**2) / np.pi

            for n in range(2, N+1):
                int_p = pol(n+1, tau[i]) / (2*n+2) - pol(n-1, tau[i]) / (2*n-2) - (-1)**n / (n**2 - 1)
                S[i, n] = int_p / gamma

    B = S @ alpha

    return B



from re import M
import openmdao.api as om
import dymos as dm
import numpy as np

import unittest
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


class ODEComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('h', shape=nn, units='m')
        self.add_input('v', shape=nn, units='m/s')
        self.add_output('hdot', shape=nn, units='m/s')
        self.add_output('vdot', shape=nn, units='m/s**2')

        self.declare_partials(of='hdot', wrt='v', rows=np.arange(nn), cols=np.arange(nn), val=1.0)

    def compute(self, inputs, outputs):
        outputs['hdot'] = inputs['v']
        outputs['vdot'] = -9.80665


class MatrixStateCannonball(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        # z is the state vector, a nn x 2 x 2 in the form of [[x, y], [vx, vy]]
        self.add_input('z', shape=(nn, 2, 2), units=None)
        self.add_output('zdot', shape=(nn, 2, 2), units=None)

        self.declare_partials(of='zdot', wrt='z', method='cs')
        self.declare_coloring(wrt=['z'], method='cs', num_full_jacs=5, tol=1.0E-12)

    def compute(self, inputs, outputs):
        outputs['zdot'][:, 0, 0] = inputs['z'][:, 1, 0]
        outputs['zdot'][:, 0, 1] = inputs['z'][:, 1, 1]
        outputs['zdot'][:, 1, 0] = 0.0
        outputs['zdot'][:, 1, 1] = -9.81


@use_tempdirs
class TestImplicitDuration(unittest.TestCase):

    @staticmethod
    def _make_simple_problem(tx, input_duration=False, fix_duration=False, expr_end=False):
        p = om.Problem()

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ODEComp, transcription=tx)
        traj.add_phase('phase', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, units='s',
                               input_duration=input_duration, fix_duration=fix_duration)
        phase.add_state('h', rate_source='hdot', fix_initial=True, units='m', solve_segments='forward')
        phase.add_state('v', rate_source='vdot', fix_initial=True, units='m/s', solve_segments='forward')

        if not expr_end:
            phase.set_duration_balance('h', val=0.0, units='m')
        else:
            phase.set_duration_balance('pe=9.80665*h', val=0.0, units='m**2/s**2')

        phase.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
        phase.linear_solver = om.DirectSolver()

        phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

        p.driver = om.ScipyOptimizeDriver(optimizer='SLSQP')
        p.driver.declare_coloring(tol=1.0E-12)

        phase.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
        phase.linear_solver = om.DirectSolver()

        p.setup()

        phase.set_time_val(initial=0, duration=2)
        phase.set_state_val('h', [30, 0])
        phase.set_state_val('v', [0, -10])

        return p

    @staticmethod
    def _make_matrix_problem(tx, shape_error=False):
        p = om.Problem()

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=MatrixStateCannonball, transcription=tx)
        traj.add_phase('phase', phase)
        p.model.add_subsystem('traj', traj)

        phase.set_time_options(fix_initial=True, units='s')
        phase.add_state('z', rate_source='zdot', fix_initial=True, solve_segments='forward')

        index = None if shape_error else [[0], [1]]
        phase.set_duration_balance('z', val=0.0, index=index)

        phase.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
        phase.linear_solver = om.DirectSolver()

        phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

        phase.nonlinear_solver = om.NewtonSolver(solve_subsystems=True)
        phase.linear_solver = om.DirectSolver()

        p.driver = om.ScipyOptimizeDriver(optimizer='SLSQP')
        p.driver.declare_coloring(tol=1.0E-12)

        p.setup()

        p.set_val('traj.phase.t_initial', 0)
        p.set_val('traj.phase.t_duration', 7)
        p.set_val('traj.phase.states:z', phase.interp('z', [[[0, 0], [10, 10]], [[10, 0], [10, -10]]]))

        return p

    def test_implicit_duration_radau(self):
        tx = dm.Radau(num_segments=12, order=3, solve_segments=False)

        p = self._make_simple_problem(tx)

        dm.run_problem(p, run_driver=False, simulate=False)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.4735192, tolerance=1E-6)
        assert_near_equal(p.get_val('traj.phase.timeseries.h')[-1], 0.0, tolerance=1E-6)

    def test_implicit_duration_gl(self):
        tx = dm.GaussLobatto(num_segments=12, order=3, solve_segments=False)

        p = self._make_simple_problem(tx)

        dm.run_problem(p, run_driver=False, simulate=True, make_plots=True)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.4735192, tolerance=1E-6)
        assert_near_equal(p.get_val('traj.phase.timeseries.h')[-1], 0.0, tolerance=1E-6)

    def test_implicit_duration_shooting(self):
        tx = dm.ExplicitShooting(num_segments=12, order=3)

        p = self._make_simple_problem(tx)

        dm.run_problem(p, run_driver=False, simulate=False)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.4735192, tolerance=1E-6)
        assert_near_equal(p.get_val('traj.phase.timeseries.h')[-1], 0.0, tolerance=1E-6)

    def test_implicit_duration_matrix_state(self):
        tx = dm.Radau(num_segments=12, order=3, solve_segments=False)

        p = self._make_matrix_problem(tx, shape_error=False)

        dm.run_problem(p, run_driver=False, simulate=False)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.03873598, tolerance=1E-5)

    @unittest.skip("set_duration_balance/add_boundary_balance doesn't currently work with expressions.")
    @require_pyoptsparse(optimizer='IPOPT')
    def test_implicit_duration_radau_expr_condition(self):
        tx = dm.Radau(num_segments=12, order=3, solve_segments=False)

        p = self._make_simple_problem(tx, expr_end=True)

        dm.run_problem(p, run_driver=False, simulate=False)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.4735192, tolerance=1E-6)
        assert_near_equal(p.get_val('traj.phase.timeseries.h')[-1], 0.0, tolerance=1E-6)

        print((p.get_val('traj.phase.timeseries.v')[-1]) ** 2 / 2)

    @unittest.skip("set_duration_balance/add_boundary_balance doesn't currently work with expressions.")
    def test_implicit_duration_gl_expr_condition(self):
        tx = dm.GaussLobatto(num_segments=12, order=3, solve_segments=False)

        p = self._make_simple_problem(tx, expr_end=True)

        dm.run_problem(p, run_driver=False, simulate=False)

        assert_near_equal(p.get_val('traj.phase.timeseries.time')[-1], 2.4735192, tolerance=1E-6)
        assert_near_equal(p.get_val('traj.phase.timeseries.h')[-1], 0.0, tolerance=1E-6)


if __name__ == '__main__':
    unittest.main()
